"""Provides handling of TCC zones (heating and DHW)."""

# TODO: extend set_mode() for non-evohome modes (e.g. "VacationHold")

from __future__ import annotations

import json
from datetime import UTC, datetime as dt, timedelta as td
from functools import cached_property
from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Final

from _evohome.helpers import as_local_time, camel_to_snake

from . import exceptions as exc
from .const import (
    API_STRFTIME,
    SZ_ALLOWED_SETPOINT_MODES,
    SZ_DAILY_SCHEDULES,
    SZ_FAULT_TYPE,
    SZ_IS_AVAILABLE,
    SZ_MAX_HEAT_SETPOINT,
    SZ_MIN_HEAT_SETPOINT,
    SZ_MODEL_TYPE,
    SZ_NAME,
    SZ_SCHEDULE_CAPABILITIES,
    SZ_SETPOINT_CAPABILITIES,
    SZ_SETPOINT_MODE,
    SZ_SETPOINT_STATUS,
    SZ_SINCE,
    SZ_TARGET_HEAT_TEMPERATURE,
    SZ_TEMPERATURE,
    SZ_TEMPERATURE_STATUS,
    SZ_ZONE_ID,
    SZ_ZONE_TYPE,
)
from .schemas import TccSetZonModeT, ZoneMode, factory_zon_schedule, factory_zon_status
from .schemas.const import (
    S2_HEAT_SETPOINT_VALUE,
    S2_SETPOINT_MODE,
    S2_TIME_UNTIL,
    EntityType,
    ZoneModelType,
    ZoneType,
)
from .schemas.schedule import DayOfWeek

if TYPE_CHECKING:
    import logging
    from datetime import tzinfo

    import voluptuous as vol

    from . import ControlSystem, Location
    from .auth import Auth
    from .schemas.typedefs import (
        DailySchedulesT,
        DayOfWeekT,
        DayOfWeekZoneT,
        EvoActiveFaultResponseT,
        EvoDhwConfigEntryT,
        EvoDhwStatusResponseT,
        EvoGwyConfigEntryT,
        EvoGwyStatusResponseT,
        EvoLocConfigEntryT,
        EvoLocStatusResponseT,
        EvoTcsConfigEntryT,
        EvoTcsStatusResponseT,
        EvoTemperatureStatusResponseT,
        EvoZonConfigEntryT,
        EvoZonConfigResponseT,
        EvoZonScheduleCapabilitiesResponseT,
        EvoZonSetpointCapabilitiesResponseT,
        EvoZonSetpointStatusResponseT,
        EvoZonStatusResponseT,
        SwitchpointT,
    )

    _ScheduleT = list[DayOfWeekT]
    _SwitchPoint = tuple[dt, float | str]


_ONE_DAY = td(days=1)


class EntityBase:
    _TYPE: EntityType  # e.g. "temperatureControlSystem", "domesticHotWater"

    _config: (
        EvoLocConfigEntryT
        | EvoGwyConfigEntryT
        | EvoTcsConfigEntryT
        | EvoZonConfigEntryT
        | EvoDhwConfigEntryT
    )

    _status: (
        EvoDhwStatusResponseT
        | EvoGwyStatusResponseT
        | EvoLocStatusResponseT
        | EvoTcsStatusResponseT
        | EvoZonStatusResponseT
        | None
    )

    def __init__(self, entity_id: str, auth: Auth, logger: logging.Logger) -> None:
        self._id: Final = entity_id

        self._auth = auth
        self._logger = logger

    def __str__(self) -> str:
        """Return a string representation of the entity."""
        return f"{self.__class__.__name__}(id='{self._id}')"

    # Config attrs...

    @cached_property
    def id(self) -> str:
        return self._id

    @property
    def config(
        self,
    ) -> (
        EvoLocConfigEntryT
        | EvoGwyConfigEntryT
        | EvoTcsConfigEntryT
        | EvoZonConfigEntryT
        | EvoDhwConfigEntryT
    ):
        """Return the latest config of the entity."""
        return self._config

    # Status (state) attrs & methods...

    @property
    def status(
        self,
    ) -> (
        EvoLocStatusResponseT
        | EvoGwyStatusResponseT
        | EvoTcsStatusResponseT
        | EvoZonStatusResponseT
        | EvoDhwStatusResponseT
    ):
        """Return the latest status of the entity."""
        if self._status is None:
            raise exc.InvalidStatusError(
                "No status available (have not invoked Location.update()?)"
            )
        return self._status


class ActiveFaultsBase(EntityBase):
    """Provide the base for active faults."""

    location: Location  # used to get tzinfo

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self._active_faults: list[EvoActiveFaultResponseT] = []
        self._last_logged: dict[str, dt] = {}  # OK to use a tz=UTC datetimes

    # Status (state) attrs & methods...

    def _update_faults(
        self,
        active_faults: list[EvoActiveFaultResponseT],
    ) -> None:
        """Maintain self._active_faults list and self._last_logged dict."""

        def hash_(fault: EvoActiveFaultResponseT) -> str:
            return f"{fault[SZ_SINCE]}_{fault[SZ_FAULT_TYPE]}"

        def log_as_active(fault: EvoActiveFaultResponseT) -> None:
            self._logger.warning(
                f"{self}: Active fault: {fault[SZ_SINCE]} {fault[SZ_FAULT_TYPE]}"
            )
            self._last_logged[hash_(fault)] = dt.now(tz=UTC)  # aware dtm not required

        def log_as_resolved(fault: EvoActiveFaultResponseT) -> None:
            self._logger.info(
                f"{self}: Fault cleared: {fault[SZ_SINCE]} {fault[SZ_FAULT_TYPE]}"
            )
            del self._last_logged[hash_(fault)]

        # Remove resolved (non-active) faults
        for fault in [f for f in self._active_faults if f not in active_faults]:
            log_as_resolved(fault)
            self._active_faults.remove(fault)

        # Add new (active) faults
        for fault in [f for f in active_faults if f not in self._active_faults]:
            log_as_active(fault)
            self._active_faults.append(fault)

        # Re-log active faults if necessary
        for fault in self._active_faults:
            if dt.now(tz=UTC) - self._last_logged[hash_(fault)] > _ONE_DAY:
                log_as_active(fault)

    @property
    def active_faults(self) -> tuple[EvoActiveFaultResponseT, ...]:
        """
        "activeFaults": [
            {
            "faultType": "GatewayCommunicationLost",
            "since": "2023-05-04T18:47:36.7727046"
            }
        ]
        """

        return tuple(self._active_faults)


def _dt_to_dow_and_tod(dtm: dt, tzinfo: tzinfo) -> tuple[DayOfWeek, str]:
    """Return a pair of strings representing the local day of week and time of day."""
    dtm = as_local_time(dtm, tzinfo)
    return dtm.strftime("%A"), dtm.strftime("%H:%M")  # type: ignore[return-value]


def _find_switchpoints(
    schedule: _ScheduleT,
    day_of_week: DayOfWeek,
    time_of_day: str,
) -> tuple[SwitchpointT, int, SwitchpointT, int]:
    """Find this/next switchpoints for a given day of week and time of day."""

    # TODO: schedule can be [], i.e. response was {'DailySchedules': []}

    # assumes 1+ switchpoint per day, which could be this_sp or next_sp only

    try:
        day_idx = list(DayOfWeek).index(day_of_week)
    except ValueError as err:
        raise TypeError(f"Invalid parameter: {day_of_week}") from err

    this_sp: SwitchpointT | None = None
    next_sp: SwitchpointT | None = None

    this_offset = 0
    next_offset = 0

    # Check the switchpoints of the given day of week
    for sp in schedule[day_idx]["switchpoints"]:
        if sp["time_of_day"] <= time_of_day:
            this_sp = sp
            continue

        if sp["time_of_day"] > time_of_day:
            if this_sp is None:
                this_sp = schedule[(day_idx + 6) % 7]["switchpoints"][-1]
                this_offset = -1

            next_sp = sp
            break

    else:
        assert this_sp is not None  # mypy

        if next_sp is None:
            next_sp = schedule[(day_idx + 1) % 7]["switchpoints"][0]
            next_offset = +1

    return this_sp, this_offset, next_sp, next_offset


class _ScheduleBase(EntityBase):
    """Provide the base for temperatureZone / domesticHotWater Zones."""

    SCH_SCHEDULE: vol.Schema

    _schedule: _ScheduleT | None

    _this_switchpoint: _SwitchPoint  # is float for zones...
    _next_switchpoint: _SwitchPoint  # and str for DHW

    location: Location  # used to get tzinfo

    # Status (state) attrs & methods...

    @property
    def schedule(self) -> _ScheduleT:
        """Return the schedule (assumes it is current)."""

        if not self._schedule:
            raise exc.InvalidScheduleError(f"{self}: No Schedule, or is invalid")

        return self._schedule

    @property
    def this_switchpoint(self) -> _SwitchPoint:
        """Return the start datetime and setpoint of the current switchpoint."""

        if not self._schedule:
            raise exc.InvalidScheduleError(f"{self}: No Schedule, or is invalid")

        if self._next_switchpoint[0] > (dt_now := dt.now(tz=UTC)):
            return self._this_switchpoint

        self._this_switchpoint, self._next_switchpoint = self._find_switchpoints(dt_now)
        return self._this_switchpoint

    @property
    def next_switchpoint(self) -> _SwitchPoint:
        """Return the start datetime and setpoint of the next switchpoint."""

        if not self._schedule:
            raise exc.InvalidScheduleError(f"{self}: No Schedule, or is invalid")

        if self._next_switchpoint[0] > (dt_now := dt.now(tz=UTC)):
            return self._next_switchpoint

        self._this_switchpoint, self._next_switchpoint = self._find_switchpoints(dt_now)

        return self._next_switchpoint

    async def get_schedule(self) -> _ScheduleT:
        """Get the schedule for this DHW/zone object."""

        self._logger.debug(f"{self}: Getting schedule...")

        try:
            schedule: DailySchedulesT = await self._auth.get(
                f"{self._TYPE}/{self.id}/schedule", schema=self.SCH_SCHEDULE
            )  # type: ignore[assignment]

        except exc.ApiRequestFailedError as err:
            if err.status == HTTPStatus.BAD_REQUEST:  # 400
                raise exc.InvalidScheduleError(
                    f"{self}: No Schedule, or is invalid"
                ) from err
            raise exc.ApiRequestFailedError(f"{self}: Unexpected error") from err

        self._schedule = schedule[SZ_DAILY_SCHEDULES]

        self._this_switchpoint, self._next_switchpoint = self._find_switchpoints(
            dt.now(tz=UTC)
        )

        return self._schedule

    def _find_switchpoints(self, dtm: dt) -> tuple[_SwitchPoint, _SwitchPoint]:
        """Find the current (this) and next switchpoints for a given datetime.

        FYI: HA has traditonally exposed (as an extended_state_attr):
        {
            "this_sp_from": "2024-07-10T08:00:00+01:00",
            "this_sp_temp": 16.0,
            "next_sp_from": "2024-07-10T22:10:00+01:00",
            "next_sp_temp": 18.6,
        }
        """

        dtm = as_local_time(dtm, self.location.tzinfo)

        this_sp, this_offset, next_sp, next_offset = _find_switchpoints(
            self.schedule, *_dt_to_dow_and_tod(dtm, self.location.tzinfo)
        )

        this_tod = dt.strptime(this_sp["time_of_day"], "%H:%M:00").time()  # noqa: DTZ007
        next_tod = dt.strptime(next_sp["time_of_day"], "%H:%M:00").time()  # noqa: DTZ007

        this_dtm = dt.combine(dtm + td(days=this_offset), this_tod)
        next_dtm = dt.combine(dtm + td(days=next_offset), next_tod)

        # either "dhw_state" (str) or "heat_setpoint" (float) _will_ be present...
        this_val = this_sp.get("dhw_state") or this_sp["heat_setpoint"]  # pyright: ignore[reportTypedDictNotRequiredAccess]
        next_val = next_sp.get("dhw_state") or next_sp["heat_setpoint"]  # pyright: ignore[reportTypedDictNotRequiredAccess]

        return (
            (this_dtm.replace(tzinfo=self.location.tzinfo), this_val),
            (next_dtm.replace(tzinfo=self.location.tzinfo), next_val),
        )

    async def set_schedule(
        self,
        schedule: _ScheduleT | str,
    ) -> None:
        """Set the schedule for this DHW/zone object."""

        self._logger.debug(f"{self}: Setting schedule...")

        if isinstance(schedule, list):
            try:
                json.dumps(schedule)
            except (OverflowError, TypeError, ValueError) as err:
                raise exc.BadScheduleUploadedError(
                    f"{self}: Invalid schedule: {err}"
                ) from err

        elif isinstance(schedule, str):
            try:
                schedule = json.loads(schedule)
            except json.JSONDecodeError as err:
                raise exc.BadScheduleUploadedError(
                    f"{self}: Invalid schedule: {err}"
                ) from err

            assert isinstance(schedule, list)  # mypy

        else:
            raise exc.BadScheduleUploadedError(
                f"{self}: Invalid schedule: {type(schedule)} is not JSON serializable"
            )

        _ = await self._auth.put(
            f"{self._TYPE}/{self.id}/schedule",
            json={"daily_schedules": schedule},
            schema=self.SCH_SCHEDULE,
        )

        # TODO: check the status of the task

        self._schedule = schedule


class _ZoneBase(_ScheduleBase, ActiveFaultsBase, EntityBase):
    """Provide the base for temperatureZone / domesticHotWater Zones."""

    SCH_STATUS: vol.Schema

    _status: EvoDhwStatusResponseT | EvoZonStatusResponseT | None

    def __init__(self, entity_id: str, tcs: ControlSystem) -> None:
        super().__init__(entity_id, tcs._auth, tcs._logger)

        self.location = tcs.location
        self.tcs = tcs

    # Status (state) attrs & methods...

    async def _get_status(self) -> EvoDhwStatusResponseT | EvoZonStatusResponseT:
        """Get the latest state of the DHW/zone and update its status attrs.

        It is more efficient to call Location.update() as all descendants are updated
        with a single GET. Returns the raw JSON of the latest state.
        """

        status: EvoDhwStatusResponseT | EvoZonStatusResponseT = await self._auth.get(  # type: ignore[assignment]
            f"{self._TYPE}/{self.id}/status", schema=self.SCH_STATUS
        )

        self._update_status(status)
        return status

    def _update_status(
        self, status: EvoDhwStatusResponseT | EvoZonStatusResponseT
    ) -> None:
        """Update the DHW/ZON's status."""

        self._update_faults(status["active_faults"])
        self._status = status

    @property
    def temperature_status(self) -> EvoTemperatureStatusResponseT:
        """
        "temperatureStatus": {
            "temperature": 20.0,
            "isAvailable": true
        }
        """

        if self._status is None:
            raise exc.InvalidStatusError(f"{self} has no state, has it been fetched?")
        return self._status[SZ_TEMPERATURE_STATUS]

    @property  # a convenience attr
    def temperature(self) -> float | None:
        if not (status := self.temperature_status) or not status[SZ_IS_AVAILABLE]:
            return None

        assert SZ_TEMPERATURE in status  # mypy hint
        return status[SZ_TEMPERATURE]


class Zone(_ZoneBase):
    """Instance of a TCS's heating zone (temperatureZone)."""

    _TYPE = EntityType.ZON

    SCH_SCHEDULE: vol.Schema = factory_zon_schedule(camel_to_snake)
    SCH_STATUS: vol.Schema = factory_zon_status(camel_to_snake)

    def __init__(self, tcs: ControlSystem, config: EvoZonConfigResponseT) -> None:
        super().__init__(config[SZ_ZONE_ID], tcs)

        self._config: Final[EvoZonConfigResponseT] = config  # type: ignore[misc]
        self._status: EvoZonStatusResponseT | None = None

        self._schedule: list[DayOfWeekZoneT] | None = None  # type: ignore[assignment]

        if not self.model or self.model == ZoneModelType.UNKNOWN:
            raise exc.InvalidConfigError(
                f"{self}: Invalid model type '{self.model}' (is it a ghost zone?)"
            )
        if not self.type or self.type == ZoneType.UNKNOWN:
            raise exc.InvalidConfigError(
                f"{self}: Invalid zone type '{self.type}' (is it a ghost zone?)"
            )

        if self.model not in ZoneModelType:
            self._logger.warning("%s: Unknown model type '%s' (YMMV)", self, self.model)
        if self.type not in ZoneType:
            self._logger.warning("%s: Unknown zone type '%s' (YMMV)", self, self.type)

    @property
    def config(self) -> EvoZonConfigEntryT:
        """Return the latest config of the entity."""
        return self._config

    @property
    def status(self) -> EvoZonStatusResponseT:
        """Return the latest status of the entity."""
        return super().status  # type: ignore[return-value]

    # Config attrs...

    @cached_property
    def model(self) -> ZoneModelType:
        return self._config[SZ_MODEL_TYPE]

    @property
    def name(self) -> str:
        if self._status is not None:
            return self._status[SZ_NAME]
        return self._config[SZ_NAME]

    @cached_property
    def type(self) -> ZoneType:
        return self._config[SZ_ZONE_TYPE]

    @cached_property
    def schedule_capabilities(self) -> EvoZonScheduleCapabilitiesResponseT:
        """
        "scheduleCapabilities": {
            "maxSwitchpointsPerDay": 6,
            "minSwitchpointsPerDay": 1,
            "timingResolution": "00:10:00",
            "setpointValueResolution": 0.5
        }
        """

        return self._config[SZ_SCHEDULE_CAPABILITIES]

    @cached_property
    def setpoint_capabilities(self) -> EvoZonSetpointCapabilitiesResponseT:
        """
        "setpointCapabilities": {
            "maxHeatSetpoint": 35.0,
            "minHeatSetpoint": 5.0,
            "valueResolution": 0.5,
            "canControlHeat": true,
            "canControlCool": false,
            "allowedSetpointModes": ["PermanentOverride", "FollowSchedule", "TemporaryOverride"],
            "maxDuration": "1.00:00:00",
            "timingResolution": "00:10:00"
        }
        """

        return self._config[SZ_SETPOINT_CAPABILITIES]

    @cached_property
    def allowed_modes(self) -> tuple[ZoneMode, ...]:
        return tuple(self.setpoint_capabilities[SZ_ALLOWED_SETPOINT_MODES])

    @property  # convenience attr
    def max_heat_setpoint(self) -> float:
        # consider: if not self.setpoint_capabilities["can_control_heat"]: return None
        return self.setpoint_capabilities[SZ_MAX_HEAT_SETPOINT]

    @property  # convenience attr
    def min_heat_setpoint(self) -> float:
        # consider: if not self.setpoint_capabilities["can_control_heat"]: return None
        return self.setpoint_capabilities[SZ_MIN_HEAT_SETPOINT]

    # Status (state) attrs & methods...

    @property
    def setpoint_status(self) -> EvoZonSetpointStatusResponseT:
        """
        "setpointStatus": {
            "targetHeatTemperature": 17.0,
            "setpointMode": "FollowSchedule"
        }
        "setpointStatus": {
            "targetHeatTemperature": 20.5,
            "setpointMode": "TemporaryOverride",
            "until": "2023-11-30T22:10:00Z"
        }
        """

        if self._status is None:
            raise exc.InvalidStatusError(f"{self} has no state, has it been fetched?")
        return self._status[SZ_SETPOINT_STATUS]

    @property
    def mode(self) -> ZoneMode:
        return self.setpoint_status[SZ_SETPOINT_MODE]

    @property
    def target_heat_temperature(self) -> float:
        return self.setpoint_status[SZ_TARGET_HEAT_TEMPERATURE]

    @property
    def until(self) -> dt | None:
        if (until := self.setpoint_status.get("until")) is None:
            return None
        return as_local_time(until, self.location.tzinfo)

    async def _set_mode(self, mode: TccSetZonModeT) -> None:
        """Set the zone mode (heating only, a cooling is not exposed by the API)."""

        # Issue a warning if we fail some basic sanity checks...
        if mode[S2_SETPOINT_MODE] not in self.allowed_modes:
            self._logger.warning(
                f"{self}: Unsupported/unknown {S2_SETPOINT_MODE}: {mode}"
            )

        if (temp := mode.get(S2_HEAT_SETPOINT_VALUE)) is None:
            if mode[S2_SETPOINT_MODE] != ZoneMode.FOLLOW_SCHEDULE:
                self._logger.warning(
                    f"{self}: Missing {S2_HEAT_SETPOINT_VALUE}: {mode}"
                )

        elif self.min_heat_setpoint > temp > self.max_heat_setpoint:
            self._logger.warning(
                f"{self}: Unsupported/invalid {S2_HEAT_SETPOINT_VALUE}: {mode}"
            )

        # Call the API...
        await self._auth.put(f"{self._TYPE}/{self.id}/heatSetpoint", json=dict(mode))

    async def reset(self) -> None:
        """Cancel any override and allow the zone to follow its schedule"""
        await self._set_mode({S2_SETPOINT_MODE: ZoneMode.FOLLOW_SCHEDULE})

    # NOTE: no provision for cooling (not supported by API)
    async def set_temperature(  # aka. set_mode()
        self, temperature: float, /, *, until: dt | None = None
    ) -> None:
        """Set the temperature of the given zone (no provision for cooling)."""

        mode: TccSetZonModeT

        if until is None:  # NOTE: beware that these may be case-sensitive
            mode = {
                S2_SETPOINT_MODE: ZoneMode.PERMANENT_OVERRIDE,
                S2_HEAT_SETPOINT_VALUE: temperature,
            }
        else:
            mode = {  # NOTE: TIME_UNTIL, not UNTIL_TIME
                S2_SETPOINT_MODE: ZoneMode.TEMPORARY_OVERRIDE,
                S2_HEAT_SETPOINT_VALUE: temperature,
                S2_TIME_UNTIL: until.strftime(API_STRFTIME),
            }

        await self._set_mode(mode)

    # NOTE: this wrapper exists only for typing purposes
    async def get_schedule(self) -> list[DayOfWeekZoneT]:  # type: ignore[override]
        """Get the schedule for this heating zone."""
        return await super().get_schedule()  # type: ignore[return-value]

    # NOTE: this wrapper exists only for typing purposes
    async def set_schedule(self, schedule: list[DayOfWeekZoneT] | str) -> None:  # type: ignore[override]
        """Set the schedule for this heating zone."""
        await super().set_schedule(schedule)  # type: ignore[arg-type]



"""evohomeasync schema - shared types (WIP)."""

from __future__ import annotations

from typing import Literal, NotRequired, TypedDict

from .config import (  # noqa: TC001
    DhwState,
    LocationType,
    SystemMode,
    TcsModelType,
    ZoneMode,
    ZoneModelType,
    ZoneType,
)


class TccAuthTokensResponseT(TypedDict):
    """Response to POST /Auth/OAuth/Token."""

    access_token: str
    expires_in: int
    scope: str
    refresh_token: str
    token_type: str


class EvoAuthTokensDictT(TypedDict):
    access_token: str
    access_token_expires: str  # dt.isoformat()
    refresh_token: str


#######################################################################################
# GET Entity Info/Config...
# NOTE: dicts are not completely typed, but all referenced keys should be present


# GET /accountInfo
class EvoUsrConfigResponseT(TypedDict):
    """Response to GET /accountInfo."""

    user_id: str


# GET /location/installationInfo?userId={user_id}&include... returns list of these dicts
class EvoLocConfigResponseT(TypedDict):
    """Response to GET /locations?userId={user_id}&allData=True

    The response is a list of these dicts.
    """

    location_info: EvoLocConfigEntryT
    gateways: list[EvoGwyConfigResponseT]


class EvoLocConfigEntryT(TypedDict):
    """Location configuration information."""

    location_id: str
    name: str
    street_address: str
    city: str
    state: str
    country: str
    postcode: str
    type: str
    location_type: LocationType
    use_daylight_save_switching: bool
    time_zone: EvoTimeZoneInfoT
    location_owner: EvoLocationOwnerInfoT


class EvoTimeZoneInfoT(TypedDict):
    """Time zone information."""

    time_zone_id: str
    display_name: str
    offset_minutes: int
    current_offset_minutes: int
    supports_daylight_saving: bool


class EvoLocationOwnerInfoT(TypedDict):
    user_id: str
    username: str
    firstname: str
    lastname: str


class EvoGwyConfigResponseT(TypedDict):
    gateway_info: EvoGwyConfigEntryT
    temperature_control_systems: list[EvoTcsConfigResponseT]


class EvoGwyConfigEntryT(TypedDict):
    gateway_id: str
    mac: str
    crc: str
    is_wi_fi: bool


class EvoTcsConfigEntryT(TypedDict):
    system_id: str
    model_type: TcsModelType
    allowed_system_modes: list[EvoAllowedSystemModesResponseT]


class EvoAllowedSystemModesResponseT(TypedDict):
    system_mode: SystemMode
    can_be_permanent: Literal[True]
    can_be_temporary: bool
    max_duration: NotRequired[str]  # when can_be_temporary is True
    timing_resolution: NotRequired[str]  # when can_be_temporary is True
    timing_mode: NotRequired[str]  # when can_be_temporary is True


class EvoTcsConfigResponseT(EvoTcsConfigEntryT):
    # system_id: str
    # model_type: str
    # allowed_system_modes: list[dict[str, Any]]
    zones: list[EvoZonConfigResponseT]
    dhw: EvoDhwConfigResponseT


class EvoZonConfigResponseT(TypedDict):
    zone_id: str
    model_type: ZoneModelType
    name: str
    setpoint_capabilities: EvoZonSetpointCapabilitiesResponseT
    schedule_capabilities: EvoZonScheduleCapabilitiesResponseT
    zone_type: ZoneType
    allowed_fan_modes: list[str]


class EvoZonScheduleCapabilitiesResponseT(TypedDict):
    pass


class EvoZonSetpointCapabilitiesResponseT(TypedDict):
    allowed_setpoint_modes: list[ZoneMode]
    can_control_cool: bool
    can_control_heat: bool
    max_heat_setpoint: float
    min_heat_setpoint: float
    value_resolution: float
    max_duration: str
    timing_resolution: str


class EvoZonConfigEntryT(EvoZonConfigResponseT):
    pass


class EvoDhwConfigResponseT(TypedDict):
    dhw_id: str
    schedule_capabilities_response: EvoDhwScheduleCapabilitiesResponseT
    dhw_state_capabilities_response: EvoDhwStateCapabilitiesResponseT


class EvoDhwScheduleCapabilitiesResponseT(TypedDict):
    pass


class EvoDhwStateCapabilitiesResponseT(TypedDict):
    allowed_states: list[DhwState]
    allowed_modes: list[ZoneMode]
    max_duration: str
    timing_resolution: str


class EvoDhwConfigEntryT(EvoDhwConfigResponseT):
    pass


#######################################################################################
# GET Entity Status...
# NOTE: dicts are not completely typed, but all referenced keys should be present


# GET /location/{loc_id}/status?include... returns this dict
class EvoLocStatusResponseT(TypedDict):
    """Response to /location/{loc_id}/status?includeTemperatureControlSystems=True

    The response is a dict of of a single location.
    """

    location_id: str
    gateways: list[EvoGwyStatusResponseT]


class EvoGwyStatusResponseT(TypedDict):
    gateway_id: str
    active_faults: list[EvoActiveFaultResponseT]
    temperature_control_systems: list[EvoTcsStatusResponseT]


class EvoActiveFaultResponseT(TypedDict):
    fault_type: str
    since: str


class EvoTcsStatusResponseT(TypedDict):
    system_id: str
    active_faults: list[EvoActiveFaultResponseT]
    system_mode_status: EvoSystemModeStatusResponseT
    zones: list[EvoZonStatusResponseT]
    dhw: NotRequired[EvoDhwStatusResponseT]


class EvoSystemModeStatusResponseT(TypedDict):
    mode: SystemMode
    is_permanent: bool
    time_until: NotRequired[str]


class EvoZonStatusResponseT(TypedDict):
    zone_id: str
    active_faults: list[EvoActiveFaultResponseT]
    setpoint_status: EvoZonSetpointStatusResponseT
    temperature_status: EvoTemperatureStatusResponseT
    name: str


class EvoZonSetpointStatusResponseT(TypedDict):
    setpoint_mode: ZoneMode
    target_heat_temperature: int
    until: NotRequired[str]


class EvoTemperatureStatusResponseT(TypedDict):
    is_available: bool
    temperature: NotRequired[int]


class EvoDhwStatusResponseT(TypedDict):
    dhw_id: str
    active_faults: list[EvoActiveFaultResponseT]
    state_status: EvoDhwStateStatusResponseT
    temperature_status: EvoTemperatureStatusResponseT


class EvoDhwStateStatusResponseT(TypedDict):
    mode: ZoneMode
    state: DhwState
    until: NotRequired[str]


#######################################################################################
# PUT TCS/Zone/DHW State...
#


# PUT /domesticHotWater/{dhw_id}/state
class EvoSetDhwStateT(TypedDict):
    mode: ZoneMode
    state: NotRequired[DhwState]  # required by override modes
    until_time: NotRequired[str]  # required by TemporaryOverride


# PUT /temperatureControlSystem/{tcs_id}/mode
class EvoSetSystemModeT(TypedDict):
    system_mode: SystemMode
    permanent: bool
    time_until: NotRequired[str]  # required by TemporaryOverride


# PUT /temperatureZone/{zon_id}/heatSetpoint
class EvoSetZoneHeatSetpointT(TypedDict):
    setpoint_mode: ZoneMode
    heat_setpoint_value: NotRequired[float]  # required by override modes
    time_until: NotRequired[str]  # required by TemporaryOverride


#######################################################################################
# GET/PUT Zone/DHW Schedules...
#


class SwitchpointDhwT(TypedDict):
    dhw_state: DhwState
    time_of_day: str


class DayOfWeekDhwT(TypedDict):
    day_of_week: str
    switchpoints: list[SwitchpointDhwT]


class DailySchedulesDhwT(TypedDict):
    daily_schedules: list[DayOfWeekDhwT]


# for export/import to/from file
class EvoScheduleDhwT(DailySchedulesDhwT):
    dhw_id: str
    name: NotRequired[str]


#


class SwitchpointZoneT(TypedDict):
    heat_setpoint: float
    time_of_day: str


class DayOfWeekZoneT(TypedDict):
    day_of_week: str
    switchpoints: list[SwitchpointZoneT]


class DailySchedulesZoneT(TypedDict):
    daily_schedules: list[DayOfWeekZoneT]


# for export/import to/from file
class EvoScheduleZoneT(DailySchedulesZoneT):
    zone_id: str
    name: NotRequired[str]


#


class SwitchpointT(TypedDict):
    time_of_day: str
    dhw_state: NotRequired[str]  # mutex with heat_setpoint
    heat_setpoint: NotRequired[float]


class DayOfWeekT(TypedDict):
    day_of_week: str
    switchpoints: list[SwitchpointT]


class DailySchedulesT(TypedDict):
    daily_schedules: list[DayOfWeekT]


# for export/import to/from file
class EvoScheduleT(DailySchedulesT):
    dhw_id: NotRequired[str]  # exactly one of these two IDs will be present
    zone_id: NotRequired[str]
    name: NotRequired[str]  # would normally be present, but be OK if not



"""evohomeasync provides an async client for the v0 Resideo TCC API."""

from __future__ import annotations

import json
import logging
from abc import ABC, abstractmethod
from functools import cached_property
from http import HTTPMethod, HTTPStatus
from typing import TYPE_CHECKING, Any, Final

import aiohttp
import voluptuous as vol

from . import exceptions as exc
from .const import ERR_MSG_LOOKUP_BASE, HINT_CHECK_NETWORK, HOSTNAME
from .helpers import (
    convert_keys_to_camel_case,
    convert_keys_to_snake_case,
    redact_secrets,
)

if TYPE_CHECKING:
    from aiohttp.typedefs import StrOrURL


async def _payload(r: aiohttp.ClientResponse | None) -> str | None:
    if r is None:
        return None

    try:
        if r.content_type == "application/json":
            return json.dumps(await r.json())
        if r.content_type == "text/plain":
            return await r.text()
        return await r.text()  # text/html?

    except aiohttp.ClientPayloadError:
        return None
    except aiohttp.ClientError:
        return None


class AbstractAuth(ABC):
    """A class to provide to access the Resideo TCC API."""

    _url_base: StrOrURL

    def __init__(
        self,
        websession: aiohttp.ClientSession,
        /,
        *,
        logger: logging.Logger,
        _hostname: str | None = None,
    ) -> None:
        """A class for interacting with the Resideo TCC API."""

        self.websession: Final = websession

        self._logger: Final = logger
        self._hostname: Final = _hostname or HOSTNAME

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"{self.__class__.__name__}(base='{self.url_base}')"

    @cached_property
    def hostname(self) -> str:
        """Return the hostname used for GET/PUT requests."""
        return self._hostname

    @property
    def url_base(self) -> StrOrURL:
        """Return the URL base used for GET/PUT requests."""
        return self._url_base

    async def get(
        self, url: StrOrURL, /, schema: vol.Schema | None = None
    ) -> dict[str, Any] | list[dict[str, Any]]:
        """Call the vendor's TCC API with a GET.

        Optionally checks the response JSON against the expected schema and logs a
        debug message if it doesn't match.
        """

        response = await self.request(HTTPMethod.GET, url)

        if schema:
            try:
                response = schema(response)
            except vol.Invalid as err:
                self._logger.debug(f"GET {url}: payload may be invalid: {err}")

        return response  # type: ignore[return-value]

    async def put(
        self,
        url: StrOrURL,
        /,
        json: dict[str, Any],
        *,
        schema: vol.Schema | None = None,
    ) -> dict[str, Any] | list[dict[str, Any]]:  # NOTE: not _EvoSchemaT
        """Call the vendor's API with a PUT.

        Optionally checks the payload JSON against the expected schema and logs a
        debug message if it doesn't match.
        """

        if schema:
            try:
                schema(json)
            except vol.Invalid as err:
                self._logger.debug(f"PUT {url}: payload may be invalid: {err}")

        return await self.request(HTTPMethod.PUT, url, json=json)  # type: ignore[return-value]

    async def request(
        self, method: HTTPMethod, url: StrOrURL, /, **kwargs: Any
    ) -> dict[str, Any] | list[dict[str, Any]] | str | None:
        """Make a request to the vendor's TCC RESTful API.

        Converts keys to/from snake_case as required.
        """

        if method == HTTPMethod.PUT and "json" in kwargs:
            kwargs["json"] = convert_keys_to_camel_case(kwargs["json"])

        try:
            response = await self._make_request(method, url, **kwargs)
        except exc.ApiRequestFailedError as err:
            if err.status != HTTPStatus.UNAUTHORIZED:  # 401
                # leave it up to higher layers to handle 401s as they can either be
                # - authentication errors: bad access_token, bad session_id
                # - authorization errors:  bad URL (e.g. no access to that loc_id)
                self._logger.debug(
                    f"The access_token/session_id may be invalid (it shouldn't be): {err}"
                )
            raise

        if self._logger.isEnabledFor(logging.DEBUG):
            self._logger.debug(
                f"{method} {self.url_base}/{url}: {redact_secrets(response)}"
            )

        if method == HTTPMethod.GET:
            return convert_keys_to_snake_case(response)
        return response

    @abstractmethod
    async def _headers(self, headers: dict[str, str]) -> dict[str, str]:
        """Ensure the authorization header is valid.

        This could take the form of an access token, or a session id.
        """

    async def _make_request(
        self, method: HTTPMethod, url: StrOrURL, /, **kwargs: Any
    ) -> dict[str, Any] | list[dict[str, Any]]:
        """Make a GET/PUT request and return the response (a dict or a list).

        Will raise an exception if the request is not successful.
        """

        rsp: aiohttp.ClientResponse | None = None  # to prevent unbound error

        url = f"{self.url_base}/{url}"
        headers = await self._headers(kwargs.pop("headers", {}))

        try:
            rsp = await self._request(method, url, headers=headers, **kwargs)
            assert rsp is not None  # mypy hint

            await rsp.read()  # so we can use rsp.json()/rsp.text(), below
            rsp.raise_for_status()

            # can't assert content_length != 0 with aioresponses, so skip that check
            if rsp.content_type != "application/json":  # usu. "text/plain", "text/html"
                raise exc.ApiRequestFailedError(
                    f"{method} {url}: response is not JSON: {await _payload(rsp)}"
                )

            if (response := await rsp.json()) is None:  # an unanticipated edge-case
                raise exc.ApiRequestFailedError(f"{method} {url}: response is null")

        except (aiohttp.ContentTypeError, json.JSONDecodeError) as err:
            raise exc.ApiRequestFailedError(
                f"{method} {url}: response is not valid JSON: {await _payload(rsp)}"
            ) from err

        # An invalid access_token / session_id will cause a 401 and we'd need to
        # re-authenticate; unfortunately, other scenarios cause 401s (e.g. wrong
        # usr_id/loc_id in URL). So leave it up to the consumer to detect/handle
        # such 401s as they can tell if was well-known URL (e.g. without usr_id)

        except aiohttp.ClientResponseError as err:
            if hint := ERR_MSG_LOOKUP_BASE.get(err.status):
                self._logger.error(hint)  # noqa: TRY400

            msg = f"{err.status} {err.message}, response={await _payload(rsp)}"

            raise exc.ApiRequestFailedError(
                f"{method} {url}: {msg}", status=err.status
            ) from err

        except aiohttp.ClientError as err:  # e.g. ClientConnectionError
            self._logger.error(HINT_CHECK_NETWORK)  # noqa: TRY400

            raise exc.ApiRequestFailedError(
                f"{method} {url}: {err}",
            ) from err

        else:
            return response  # type: ignore[no-any-return]

        finally:
            if rsp is not None:
                rsp.release()

    async def _request(  # dev/test wrapper
        self, method: HTTPMethod, url: StrOrURL, /, **kwargs: Any
    ) -> aiohttp.ClientResponse:
        """Wrap the request to the ClientSession (useful for dev/test)."""
        return await self.websession.request(method, url, **kwargs)



"""Tests for evohome-async - validate the schemas of vendor's RESTful JSON."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from evohomeasync2.schemas import (
    TCC_GET_USR_ACCOUNT,
    TCC_GET_USR_LOCATIONS,
    factory_loc_status,
)

from .common import assert_schema
from .conftest import FIXTURES_V2

if TYPE_CHECKING:
    import pytest


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    def id_fnc(folder_path: Path) -> str:
        return folder_path.name

    folders = [
        p
        for p in Path(FIXTURES_V2).glob("*")
        if p.is_dir() and not p.name.startswith("_")
    ]
    metafunc.parametrize("folder", sorted(folders), ids=id_fnc)


def test_user_account(folder: Path) -> None:
    """Test the user account schema against the corresponding JSON."""

    assert_schema(folder, TCC_GET_USR_ACCOUNT, "user_account.json")


def test_user_locations(folder: Path) -> None:
    """Test the user locations config schema against the corresponding JSON."""

    assert_schema(folder, TCC_GET_USR_LOCATIONS, "user_locations.json")


def test_location_status(folder: Path) -> None:
    """Test the location status schema against the corresponding JSON."""

    SCH_STATUS = factory_loc_status()

    for p in Path(folder).glob("status_*.json"):
        assert_schema(folder, SCH_STATUS, p.name)



"""Tests for evohome-async - validate the instantiation of entities."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import yaml

from .common import serializable_attrs
from .conftest import FIXTURES_V0

if TYPE_CHECKING:
    import pytest
    from freezegun.api import FrozenDateTimeFactory
    from syrupy import SnapshotAssertion

    from tests.conftest import EvohomeClientv0


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    folders = [
        p
        for p in Path(FIXTURES_V0).glob("*")
        if p.is_dir() and not p.name.startswith("_")
    ]
    metafunc.parametrize(
        "fixture_folder", sorted(folders), ids=(p.name for p in sorted(folders))
    )


async def test_system_snapshot(
    evohome_v0: EvohomeClientv0,
    freezer: FrozenDateTimeFactory,
    snapshot: SnapshotAssertion,
) -> None:
    """Test the user account schema against the corresponding JSON."""

    freezer.move_to("2025-01-01T00:00:00+00:00")

    # architecture is: loc/tcs -> gwy, loc/tcs -> dhw|zon

    loc = evohome_v0.locations[0]
    assert serializable_attrs(loc) == snapshot(name="location")

    gwy = loc.gateways[0]
    assert serializable_attrs(gwy) == snapshot(name="gateway")

    # tcs = gwy.systems[0]  # V0 does *not* have loc -> gwy -> tcs
    # assert serializable_attrs(tcs) == snapshot(name="control_system")

    if dhw := loc.hotwater:
        assert serializable_attrs(dhw) == snapshot(name="hot_water")

    zones = {z.id: serializable_attrs(z) for z in loc.zones}
    assert yaml.dump(zones, indent=4) == snapshot(name="zones")



"""Validate the handling of the vendor's v2 APIs (URLs) for Schedules.

This is used to:
  a) document the RESTful API that is provided by the vendor
  b) confirm the faked server (if any) is behaving as per a)

Testing is at HTTP request layer (e.g. GET/PUT).
Everything to/from the RESTful API is in camelCase (so those schemas are used).
"""

from __future__ import annotations

from http import HTTPMethod, HTTPStatus
from typing import TYPE_CHECKING, NotRequired, TypedDict

from evohomeasync2 import schemas
from tests.const import _DBG_USE_REAL_AIOHTTP

from .common import should_fail_v2, should_work_v2, skipif_auth_failed

if TYPE_CHECKING:
    from tests.conftest import EvohomeClientv2


class SwitchpointT(TypedDict):
    timeOfDay: str
    dhwState: NotRequired[str]  # mutex with heat_setpoint
    heatSetpoint: NotRequired[float]


class DayOfWeekT(TypedDict):
    day_of_week: str
    switchpoints: list[SwitchpointT]


class DailySchedulesT(TypedDict):
    dailySchedules: list[DayOfWeekT]


async def _test_schedule_put(evo: EvohomeClientv2) -> None:
    """Test /{x._TYPE}/{x.id}/schedule"""

    schedule: DailySchedulesT  # {'dailySchedules': [...]}

    # TODO: remove .update() and use URLs only
    await evo.update()

    zone = evo.locations[0].gateways[0].systems[0].zones[0]
    url = f"{zone._TYPE}/{zone.id}/schedule"

    #
    # STEP 1: GET the current schedule
    schedule = await should_work_v2(
        evo.auth, HTTPMethod.GET, url, schema=schemas.TCC_GET_SCHEDULE
    )  # type: ignore[assignment]

    # an example of the expected response:
    """
        {
            'dailySchedules': [
                {
                    'dayOfWeek': 'Monday',
                    'switchpoints': [
                        {'heatSetpoint': 23.1, 'timeOfDay': '06:30:00'},
                        {'heatSetpoint': 18.0, 'timeOfDay': '08:00:00'},
                        {'heatSetpoint': 19.1, 'timeOfDay': '17:00:00'},
                        {'heatSetpoint': 14.9, 'timeOfDay': '21:30:00'}
                    ]
                }, {
                    'dayOfWeek': 'Tuesday',
                    'switchpoints': [{'heatSetpoint': 19.1, 'timeOfDay': '06:30:00'}, {'heatSetpoint': 18.0, 'timeOfDay': '08:00:00'}, {'heatSetpoint': 19.1, 'timeOfDay': '17:00:00'}, {'heatSetpoint': 14.9, 'timeOfDay': '21:30:00'}]
                },{
                    'dayOfWeek': 'Wednesday',
                    'switchpoints': [{'heatSetpoint': 19.1, 'timeOfDay': '06:30:00'}, {'heatSetpoint': 18.0, 'timeOfDay': '08:00:00'}, {'heatSetpoint': 19.1, 'timeOfDay': '17:00:00'}, {'heatSetpoint': 14.9, 'timeOfDay': '21:30:00'}]
                }, {
                    'dayOfWeek': 'Thursday',
                    'switchpoints': [{'heatSetpoint': 19.1, 'timeOfDay': '06:30:00'}, {'heatSetpoint': 18.0, 'timeOfDay': '08:00:00'}, {'heatSetpoint': 19.1, 'timeOfDay': '17:00:00'}, {'heatSetpoint': 14.9, 'timeOfDay': '21:30:00'}]
                }, {
                    'dayOfWeek': 'Friday',
                    'switchpoints': [{'heatSetpoint': 19.1, 'timeOfDay': '06:30:00'}, {'heatSetpoint': 18.0, 'timeOfDay': '08:00:00'}, {'heatSetpoint': 19.1, 'timeOfDay': '17:00:00'}, {'heatSetpoint': 14.9, 'timeOfDay': '21:30:00'}]
                }, {
                    'dayOfWeek': 'Saturday',
                    'switchpoints': [{'heatSetpoint': 19.1, 'timeOfDay': '07:30:00'}, {'heatSetpoint': 18.5, 'timeOfDay': '11:00:00'}, {'heatSetpoint': 19.1, 'timeOfDay': '17:00:00'}, {'heatSetpoint': 14.9, 'timeOfDay': '21:30:00'}]
                }, {
                    'dayOfWeek': 'Sunday',
                    'switchpoints': [{'heatSetpoint': 19.1, 'timeOfDay': '07:30:00'}, {'heatSetpoint': 18.5, 'timeOfDay': '11:00:00'}, {'heatSetpoint': 19.1, 'timeOfDay': '17:00:00'}, {'heatSetpoint': 14.9, 'timeOfDay': '21:30:00'}]
                }
            ]
        }
    """

    #
    # STEP 2: PUT a null schedule  # NOTE: json=None
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.PUT, url, json=None, status=HTTPStatus.BAD_REQUEST
    )

    # the expected response:
    """
        [{
            'code': 'ParameterIsMissing',
            'parameterName': 'request',
            'message': 'Parameter is missing.'
    }]
    """

    #
    # STEP 3: PUT an invalid schedule  # NOTE: json={}
    _ = await should_fail_v2(
        evo.auth, HTTPMethod.PUT, url, json={}, status=HTTPStatus.BAD_REQUEST
    )

    # the expected response:
    """
        [{
            'code': 'DayOfWeekMissing',
            'message': 'One or more days missing.'
        }]
    """

    #
    # STEP 4: PUT a valid schedule
    _ = await should_work_v2(evo.auth, HTTPMethod.PUT, url, json=dict(schedule))

    # an example of the expected response:
    """
        {'id': '840367013'}
    """


async def _test_schedule_tsk(evo: EvohomeClientv2) -> None:
    """Test /{x._TYPE}/{x.id}/schedule

    Also test commTasks?commTaskId={task_id}
    """

    schedule: DailySchedulesT  # {'dailySchedules': [...]}

    # TODO: remove .update() and use URLs only
    await evo.update()

    zone = evo.locations[0].gateways[0].systems[0].zones[0]
    url = f"{zone._TYPE}/{zone.id}/schedule"

    #
    # STEP 1: GET the current schedule
    schedule = await should_work_v2(
        evo.auth, HTTPMethod.GET, url, schema=schemas.TCC_GET_SCHEDULE
    )  # type: ignore[assignment]

    assert isinstance(schedule, dict | list)  # mypy

    if not _DBG_USE_REAL_AIOHTTP:  # TODO: faked server using old schema
        return

    #
    # STEP 2: PUT a new schedule
    temp = schedule["dailySchedules"][0]["switchpoints"][0]["heatSetpoint"]
    schedule["dailySchedules"][0]["switchpoints"][0]["heatSetpoint"] = temp + 1

    status = await should_work_v2(evo.auth, HTTPMethod.PUT, url, json=dict(schedule))

    assert isinstance(status, dict | list)  # mypy

    #
    # STEP 2: check the status of the task
    if _DBG_USE_REAL_AIOHTTP:
        task_id = status[0]["id"] if isinstance(status, list) else status["id"]

        status = await should_work_v2(
            evo.auth, HTTPMethod.GET, f"commTasks?commTaskId={task_id}"
        )
        # {'commtaskId': '840367013', 'state': 'Created'}
        # {'commtaskId': '840367013', 'state': 'Running'}
        # {'commtaskId': '840367013', 'state': 'Succeeded'}

        assert isinstance(status, dict)  # mypy  # TODO: use a SCHEMA
        assert status["commtaskId"] == task_id
        assert status["state"] in ("Created", "Running", "Succeeded")

    #
    # STEP 3: check the new schedule was effected
    schedule = await should_work_v2(
        evo.auth, HTTPMethod.GET, url, schema=schemas.TCC_GET_SCHEDULE
    )  # type: ignore[assignment]

    assert schedule["dailySchedules"][0]["switchpoints"][0]["heatSetpoint"] == temp + 1
    schedule["dailySchedules"][0]["switchpoints"][0]["heatSetpoint"] = temp

    #
    # STEP 4: PUT the original schedule back
    _ = await should_work_v2(evo.auth, HTTPMethod.PUT, url, json=dict(schedule))

    #
    # STEP 5: (optional) check the status of the task

    #
    # STEP 6: check the original schedule was effected
    schedule = await should_work_v2(
        evo.auth, HTTPMethod.GET, url, schema=schemas.TCC_GET_SCHEDULE
    )  # type: ignore[assignment]

    assert schedule["dailySchedules"][0]["switchpoints"][0]["heatSetpoint"] == temp


@skipif_auth_failed  # GET, PUT
async def test_schedule_put(evohome_v2: EvohomeClientv2) -> None:
    """Test /{x._TYPE}/{x.id}/schedule

    Does not test /commTasks?commTaskId={task_id}
    """

    await _test_schedule_put(evohome_v2)


@skipif_auth_failed  # GET, PUT
async def test_schedule_tsk(evohome_v2: EvohomeClientv2) -> None:
    """Test /{x._TYPE}/{x.id}/schedule

    Also tests /commTasks?commTaskId={task_id}
    """

    await _test_schedule_tsk(evohome_v2)



"""An async client for the v0 Resideo TCC API."""

from __future__ import annotations

from _evohome.exceptions import (
    ApiRateLimitExceededError,
    ApiRequestFailedError,
    AuthenticationFailedError,
    BadApiRequestError,
    BadApiResponseError,
    BadApiSchemaError,
    BadScheduleUploadedError,
    BadUserCredentialsError,
    ConfigError,
    EvohomeError,
    InvalidConfigError,
    InvalidScheduleError,
    InvalidStatusError,
    NoSingleTcsError,
    StatusError,
)

__all__ = [
    "ApiRateLimitExceededError",
    "ApiRequestFailedError",
    "AuthenticationFailedError",
    "BadApiRequestError",
    "BadApiResponseError",
    "BadApiSchemaError",
    "BadScheduleUploadedError",
    "BadUserCredentialsError",
    "ConfigError",
    "EvohomeError",
    "InvalidConfigError",
    "InvalidScheduleError",
    "InvalidStatusError",
    "NoSingleTcsError",
    "StatusError",
]



"""evohomeasync provides an async client for the v2 Resideo TCC API."""

from __future__ import annotations

import asyncio
import logging
from http import HTTPStatus
from typing import TYPE_CHECKING, Final
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from aiozoneinfo import async_get_time_zone

from _evohome.helpers import camel_to_snake

from . import exceptions as exc
from .auth import AbstractTokenManager, Auth
from .const import SZ_USER_ID
from .location import Location, create_location
from .schemas import factory_user_account, factory_user_locations_installation_info

if TYPE_CHECKING:
    import aiohttp

    from .control_system import ControlSystem
    from .schemas.typedefs import EvoLocConfigResponseT, EvoUsrConfigResponseT


SCH_USER_ACCOUNT: Final = factory_user_account(camel_to_snake)
SCH_USER_LOCATIONS: Final = factory_user_locations_installation_info(camel_to_snake)


_LOGGER = logging.getLogger(__name__.rpartition(".")[0])


class EvohomeClient:
    """Provide a client to access the Resideo TCC API."""

    _user_info: EvoUsrConfigResponseT | None = None
    _user_locs: list[EvoLocConfigResponseT] | None = None  # all locations of the user

    def __init__(
        self,
        token_manager: AbstractTokenManager,
        /,
        *,
        websession: aiohttp.ClientSession | None = None,
        debug: bool = False,
    ) -> None:
        """Construct the v2 EvohomeClient object."""

        self._logger = _LOGGER
        if debug:
            self._logger.setLevel(logging.DEBUG)
            self._logger.debug("Debug mode explicitly enabled via kwarg.")

        self._token_manager = token_manager
        self.auth = Auth(token_manager, websession or token_manager.websession)

        self._locations: list[Location] | None = None  # to preserve the order
        self._location_by_id: dict[str, Location] | None = None

        self._tzinfo: ZoneInfo | None = None
        self._init_tzinfo = asyncio.create_task(self._async_init_tzinfo())

    def __str__(self) -> str:
        """Return a string representation of this object."""
        return f"{self.__class__.__name__}(auth='{self.auth}')"

    async def update(
        self,
        /,
        *,
        dont_update_status: bool = False,
        _reset_config: bool = False,  # for use by test suite
    ) -> list[EvoLocConfigResponseT]:
        """Retrieve the latest state of the user's locations.

        If required (or when `_reset_config` was true), first retrieves the user
        information & the configuration of all their locations.

        There is one API call for the user info, and a second for the config of all the
        user's locations; there are additional API calls for each location's status.

        If `disable_status_update` is true, does not update the status of each location
        hierarchy (and so, does not make those additional API calls).
        """

        if _reset_config:
            self._user_info = None
            self._user_locs = None

            self._locations = None
            self._location_by_id = None

        if self._user_locs is None:
            if not self._init_tzinfo.done():
                await self._init_tzinfo

            await self._get_config(dont_update_status=dont_update_status)

        if not dont_update_status:  # don't retrieve/update status of location hierarchy
            #
            for loc in self.locations:
                await loc.update()
                #

        assert self._user_locs is not None  # mypy
        return self._user_locs

    async def _async_init_tzinfo(self) -> None:
        """Initialize timezone info without blocking the event loop."""

        # NOTE: below is an attempt to determine the local TZ of the host running this
        # client, and is not necessarily the TZ of each location known to this client;
        # locations each have their own TZ

        try:
            self._tzinfo = await async_get_time_zone("localtime")
        except ZoneInfoNotFoundError:  # e.g. on Windows
            self._tzinfo = None

    async def _get_config(
        self, /, *, dont_update_status: bool = False
    ) -> list[EvoLocConfigResponseT]:
        """Ensures the config of the user and their locations.

        If required, first retrieves the user information & installation configuration.
        """

        if self._user_info is None:  # will handle access_token rejection
            url = "userAccount"
            try:
                self._user_info = await self.auth.get(url, schema=SCH_USER_ACCOUNT)  # type: ignore[assignment]

            except exc.ApiRequestFailedError as err:  # check if 401 - bad access_token
                if err.status != HTTPStatus.UNAUTHORIZED:  # 401
                    raise

                # as the userAccount URL is open to all authenticated users, any 401 is
                # due the (albeit valid) access_token being rejected by the server

                self._logger.warning(
                    f"The access_token has been rejected (will re-authenticate): {err}"
                )

                self._token_manager._clear_access_token()
                self._user_info = await self.auth.get(url, schema=SCH_USER_ACCOUNT)  # type: ignore[assignment]

            assert self._user_info is not None  # mypy

        if self._user_locs is None:
            try:
                user_id = self._user_info[SZ_USER_ID]
            except (KeyError, TypeError) as err:
                raise exc.BadApiResponseError(
                    f"No user_id in user_info dict. Received: {self._user_info}"
                ) from err

            self._user_locs = await self.auth.get(
                f"location/installationInfo?userId={user_id}&includeTemperatureControlSystems=True",
                schema=SCH_USER_LOCATIONS,
            )  # type: ignore[assignment]

            assert self._user_locs is not None  # mypy

        if self._locations is None:
            self._locations = []
            self._location_by_id = {}

            for loc_config in self._user_locs:
                loc = await create_location(self, loc_config)
                self._locations.append(loc)
                self._location_by_id[loc.id] = loc

            # only warn once per config refresh (i.e. not on every status update)
            if not dont_update_status and (num := len(self._locations)) > 1:
                self._logger.warning(
                    f"There are {num} locations. Reduce the risk of exceeding API rate "
                    "limits by individually updating only necessary locations."
                )

        return self._user_locs

    @property
    def user_account(self) -> EvoUsrConfigResponseT:
        """Return the (config) information of the user account."""

        if not self._user_info:
            raise exc.InvalidConfigError(
                "The account information is not (yet) available"
            )

        return self._user_info

    @property
    def locations(self) -> list[Location]:
        """Return the list of location entities."""

        if not self._user_locs:
            raise exc.InvalidConfigError(
                "The installation information is not (yet) available"
            )

        return self._locations  # type: ignore[return-value]

    @property
    def location_by_id(self) -> dict[str, Location]:
        """Return the list of location entities."""

        if not self._user_locs:
            raise exc.InvalidConfigError(
                "The installation information is not (yet) available"
            )

        return self._location_by_id  # type: ignore[return-value]

    # A significant majority of users will have exactly one TCS, thus for convenience...
    @property
    def tcs(self) -> ControlSystem:
        """If there is a single TCS, return it, or raise an exception.

        The majority of users will have only one TCS.
        """

        if not (locs := self.locations) or len(locs) != 1:
            raise exc.NoSingleTcsError(
                "There is not a single location (only) for this account"
            )

        if not (gwys := locs[0].gateways) or len(gwys) != 1:
            raise exc.NoSingleTcsError(
                "There is not a single gateway (only) for this account/location"
            )

        if not (tcss := gwys[0].systems) or len(tcss) != 1:
            raise exc.NoSingleTcsError(
                "There is not a single TCS (only) for this account/location/gateway"
            )

        return tcss[0]



"""evohomeasync schema - shared constants."""

from __future__ import annotations

from enum import EnumCheck, StrEnum, verify
from typing import Final

# Various useful regex forms
REGEX_DHW_ID: Final = r"[0-9]*"
REGEX_GATEWAY_ID: Final = r"[0-9]*"
REGEX_LOCATION_ID: Final = r"[0-9]*"
REGEX_SYSTEM_ID: Final = r"[0-9]*"
REGEX_ZONE_ID: Final = r"[0-9]*"

# Datetime format used by the vendor's API
API_STRFTIME: Final = "%Y-%m-%dT%H:%M:%SZ"


# These are vendorconstants, used for keys in the vendor's schema
S2_ACTIVE_FAULTS: Final = "activeFaults"
S2_ALLOWED_FAN_MODES: Final = "allowedFanModes"
S2_ALLOWED_MODES: Final = "allowedModes"
S2_ALLOWED_SETPOINT_MODES: Final = "allowedSetpointModes"
S2_ALLOWED_STATES: Final = "allowedStates"
S2_ALLOWED_SYSTEM_MODES: Final = "allowedSystemModes"

S2_CAN_BE_PERMANENT: Final = "canBePermanent"
S2_CAN_BE_CHANGED: Final = "canBeChanged"
S2_CAN_BE_TEMPORARY: Final = "canBeTemporary"
S2_CAN_CONTROL_COOL: Final = "canControlCool"
S2_CAN_CONTROL_HEAT: Final = "canControlHeat"
S2_CITY: Final = "city"
S2_COOL_SETPOINT: Final = "coolSetpoint"
S2_COUNTRY: Final = "country"
S2_CRC: Final = "crc"
S2_CURRENT_OFFSET_MINUTES: Final = "currentOffsetMinutes"

S2_DAILY_SCHEDULES: Final = "dailySchedules"
S2_DAY_OF_WEEK: Final = "dayOfWeek"
S2_DHW: Final = "dhw"
S2_DHW_ID: Final = "dhwId"
S2_DHW_STATE: Final = "dhwState"
S2_DHW_STATE_CAPABILITIES_RESPONSE: Final = "dhwStateCapabilitiesResponse"
S2_DISPLAY_NAME: Final = "displayName"

S2_FAN_MODE: Final = "fanMode"
S2_FAN_STATUS: Final = "fanStatus"
S2_FAULT_TYPE: Final = "faultType"
S2_FIRSTNAME: Final = "firstname"

S2_GATEWAY_ID: Final = "gatewayId"
S2_GATEWAY_INFO: Final = "gatewayInfo"
S2_GATEWAYS: Final = "gateways"

S2_HEAT_SETPOINT: Final = "heatSetpoint"

S2_IS_AVAILABLE: Final = "isAvailable"
S2_IS_CANCELABLE: Final = "isCancelable"
S2_IS_CHANGEABLE: Final = "isChangeable"
S2_IS_PERMANENT: Final = "isPermanent"
S2_IS_WI_FI: Final = "isWiFi"

S2_LANGUAGE: Final = "language"
S2_LASTNAME: Final = "lastname"
S2_LOCATION_ID: Final = "locationId"
S2_LOCATION_INFO: Final = "locationInfo"
S2_LOCATION_OWNER: Final = "locationOwner"
S2_LOCATION_TYPE: Final = "locationType"

S2_MAC: Final = "mac"
S2_MAX_COOL_SETPOINT: Final = "maxCoolSetpoint"  # an extrapolation
S2_MAX_DURATION: Final = "maxDuration"
S2_MAX_HEAT_SETPOINT: Final = "maxHeatSetpoint"
S2_MAX_SWITCHPOINTS_PER_DAY: Final = "maxSwitchpointsPerDay"
S2_MIN_COOL_SETPOINT: Final = "minCoolSetpoint"
S2_MIN_DURATION: Final = "minDuration"
S2_MIN_HEAT_SETPOINT: Final = "minHeatSetpoint"
S2_MIN_SWITCHPOINTS_PER_DAY: Final = "minSwitchpointsPerDay"
S2_MODE: Final = "mode"
S2_MODEL_TYPE: Final = "modelType"

S2_NAME: Final = "name"

S2_OFFSET_MINUTES: Final = "offsetMinutes"

S2_POSTCODE: Final = "postcode"

S2_SCHEDULE_CAPABILITIES: Final = "scheduleCapabilities"
S2_SCHEDULE_CAPABILITIES_RESPONSE: Final = "scheduleCapabilitiesResponse"
S2_SETPOINT_CAPABILITIES: Final = "setpointCapabilities"
S2_SETPOINT_DEADBAND: Final = "setpointDeadband"
S2_SETPOINT_MODE: Final = "setpointMode"
S2_SETPOINT_STATUS: Final = "setpointStatus"
S2_SETPOINT_VALUE_RESOLUTION: Final = "setpointValueResolution"
S2_SINCE: Final = "since"
S2_STATE: Final = "state"
S2_STATE_STATUS: Final = "stateStatus"
S2_STREET_ADDRESS: Final = "streetAddress"
S2_SUPPORTS_DAYLIGHT_SAVING: Final = "supportsDaylightSaving"
S2_SWITCHPOINTS: Final = "switchpoints"
S2_SYSTEM_ID: Final = "systemId"
S2_SYSTEM_MODE: Final = "systemMode"
S2_SYSTEM_MODE_STATUS: Final = "systemModeStatus"

S2_TARGET_COOL_TEMPERATURE: Final = "targetCoolTemperature"  # an extrapolation
S2_TARGET_HEAT_TEMPERATURE: Final = "targetHeatTemperature"
S2_TEMPERATURE: Final = "temperature"
S2_TEMPERATURE_CONTROL_SYSTEMS: Final = "temperatureControlSystems"
S2_TEMPERATURE_STATUS: Final = "temperatureStatus"
S2_TIME_OF_DAY: Final = "timeOfDay"
S2_TIME_UNTIL: Final = "timeUntil"
S2_TIME_ZONE: Final = "timeZone"
S2_TIME_ZONE_ID: Final = "timeZoneId"
S2_TIMING_MODE: Final = "timingMode"
S2_TIMING_RESOLUTION: Final = "timingResolution"

S2_UNTIL: Final = "until"
S2_USE_DAYLIGHT_SAVE_SWITCHING: Final = "useDaylightSaveSwitching"
S2_USER_ID: Final = "userId"
S2_USERNAME: Final = "username"

S2_VACATION_HOLD_CAPABILITIES: Final = "vacationHoldCapabilities"
S2_VALUE_RESOLUTION: Final = "valueResolution"

S2_ZONE_ID: Final = "zoneId"
S2_ZONE_TYPE: Final = "zoneType"
S2_ZONES: Final = "zones"


#
# These are vendor constants, used for API calls

S2_DOMESTIC_HOT_WATER: Final = "domesticHotWater"
S2_GATEWAY: Final = "gateway"
S2_LOCATION: Final = "location"
S2_TEMPERATURE_CONTROL_SYSTEM: Final = "temperatureControlSystem"
S2_TEMPERATURE_ZONE: Final = "temperatureZone"
S2_USER_ACCOUNT: Final = "userAccount"

S2_COOL_SETPOINT_VALUE: Final = "coolSetpointValue"  # an extrapolation
S2_HEAT_SETPOINT_VALUE: Final = "heatSetpointValue"

S2_PERMANENT: Final = "permanent"
S2_UNTIL_TIME: Final = "untilTime"

# These are vendor-specific constants, used for values
S2_DURATION: Final = "Duration"
S2_HEATING_ZONE: Final = "HeatingZone"
S2_PERIOD: Final = "Period"
S2_UNKNOWN: Final = "Unknown"

#
S2_MONDAY: Final = "Monday"
S2_TUESDAY: Final = "Tuesday"
S2_WEDNESDAY: Final = "Wednesday"
S2_THURSDAY: Final = "Thursday"
S2_FRIDAY: Final = "Friday"
S2_SATURDAY: Final = "Saturday"
S2_SUNDAY: Final = "Sunday"


@verify(EnumCheck.UNIQUE)
class DayOfWeek(StrEnum):
    MONDAY = S2_MONDAY
    TUESDAY = S2_TUESDAY
    WEDNESDAY = S2_WEDNESDAY
    THURSDAY = S2_THURSDAY
    FRIDAY = S2_FRIDAY
    SATURDAY = S2_SATURDAY
    SUNDAY = S2_SUNDAY


DAYS_OF_WEEK: Final = (
    S2_MONDAY,
    S2_TUESDAY,
    S2_WEDNESDAY,
    S2_THURSDAY,
    S2_FRIDAY,
    S2_SATURDAY,
    S2_SUNDAY,
)


S2_OFF: Final = "Off"
S2_ON: Final = "On"


S2_AUTO: Final = "Auto"
S2_AUTO_WITH_ECO: Final = "AutoWithEco"
S2_AUTO_WITH_RESET: Final = "AutoWithReset"
S2_AWAY: Final = "Away"
S2_CUSTOM: Final = "Custom"
S2_DAY_OFF: Final = "DayOff"
S2_HEATING_OFF: Final = "HeatingOff"

S2_COOL: Final = "Cool"
S2_HEAT: Final = "Heat"


@verify(EnumCheck.UNIQUE)
class DhwState(StrEnum):
    OFF = S2_OFF
    ON = S2_ON


@verify(EnumCheck.UNIQUE)
class FanMode(StrEnum):
    AUTO = S2_AUTO
    ON = S2_ON


@verify(EnumCheck.UNIQUE)
class FaultType(StrEnum):  # NOTE: This list is incomplete
    SYS_B_CL = "BoilerCommunicationLost"
    SYS_C_CL = "ChValveCommunicationLost"
    DHW_A_FL = "DHWActuatorFailure"
    # W_A_CL = "DHWActuatorCommunicationLost"  # extrapolated
    DHW_S_CL = "DHWSensorCommunicationLost"
    DHW_S_FL = "DHWSensorFailure"
    DHW_S_LB = "DHWSensorLowBattery"  # extrapolated
    GWY_X_CL = "GatewayCommunicationLost"
    # S_X_LB = "TemperatureControlSystemLowBattery"  # extrapolated
    ZON_A_CL = "TempZoneActuatorCommunicationLost"
    ZON_A_LB = "TempZoneActuatorLowBattery"
    ZON_S_CL = "TempZoneSensorCommunicationLost"
    ZON_S_LB = "TempZoneSensorLowBattery"


@verify(EnumCheck.UNIQUE)
class LocationType(StrEnum):
    COMMERCIAL = "Commercial"
    AUTO_WITH_ECO = "Residential"


@verify(EnumCheck.UNIQUE)
class SystemMode(StrEnum):
    AUTO = S2_AUTO
    AUTO_WITH_ECO = S2_AUTO_WITH_ECO
    AUTO_WITH_RESET = S2_AUTO_WITH_RESET
    AWAY = S2_AWAY
    CUSTOM = S2_CUSTOM
    DAY_OFF = S2_DAY_OFF
    HEATING_OFF = S2_HEATING_OFF
    OFF = S2_OFF  # not evohome (VisionProWifiRetail)
    HEAT = S2_HEAT  # not evohome (VisionProWifiRetail)
    COOL = S2_COOL  # not evohome (VisionProWifiRetail)


@verify(EnumCheck.UNIQUE)
class EntityType(StrEnum):
    LOC = S2_LOCATION
    GWY = S2_GATEWAY
    TCS = S2_TEMPERATURE_CONTROL_SYSTEM
    ZON = S2_TEMPERATURE_ZONE
    DHW = S2_DOMESTIC_HOT_WATER


@verify(EnumCheck.UNIQUE)
class TcsModelType(StrEnum):
    EVO_TOUCH = "EvoTouch"
    FOCUS_PRO_WIFI_RETAIL = "FocusProWifiRetail"
    VISION_PRO_WIFI_RETAIL = "VisionProWifiRetail"


S2_FOLLOW_SCHEDULE = "FollowSchedule"
S2_PERMANENT_OVERRIDE = "PermanentOverride"
S2_TEMPORARY_OVERRIDE = "TemporaryOverride"
S2_VACATION_HOLD = "VacationHold"


@verify(EnumCheck.UNIQUE)
class ZoneMode(StrEnum):
    FOLLOW_SCHEDULE = S2_FOLLOW_SCHEDULE
    PERMANENT_OVERRIDE = S2_PERMANENT_OVERRIDE
    TEMPORARY_OVERRIDE = S2_TEMPORARY_OVERRIDE
    VACATION_HOLD = S2_VACATION_HOLD  # not evohome (VisionProWifiRetail)


@verify(EnumCheck.UNIQUE)
class ZoneModelType(StrEnum):
    FOCUS_PRO_WIFI_RETAIL = "FocusProWifiRetail"
    HEATING_ZONE = "HeatingZone"
    ROUND_MODULATION = "RoundModulation"
    ROUND_WIRELESS = "RoundWireless"
    UNKNOWN = S2_UNKNOWN
    VISION_PRO_WIFI_RETAIL = "VisionProWifiRetail"


S2_ELECTRIC_HEAT = "ElectricHeat"  # TODO: needs confirming
S2_MIXING_VALVE = "MixingValve"
S2_RADIATOR_ZONE = "RadiatorZone"
S2_THERMOSTAT = "Thermostat"
S2_UNDERFLOOR_HEATING = "UnderfloorHeating"
S2_ZONE_VALVES = "ZoneValves"  # is not ZoneValve
S2_ZONE_TEMPERATURE_CONTROL = "ZoneTemperatureControl"


@verify(EnumCheck.UNIQUE)
class ZoneType(StrEnum):
    MIXING_VALVE = S2_MIXING_VALVE
    RADIATOR_ZONE = S2_RADIATOR_ZONE
    THERMOSTAT = S2_THERMOSTAT
    UNDERFLOOR_HEATING = S2_UNDERFLOOR_HEATING
    UNKNOWN = S2_UNKNOWN
    ZONE_TEMPERATURE_CONTROL = S2_ZONE_TEMPERATURE_CONTROL
    ZONE_VALVES = S2_ZONE_VALVES



"""Tests for evohome-async - validate the schema of HA's debug JSON (newer ver)."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Final

from _evohome.helpers import convert_keys_to_snake_case
from evohomeasync2.location import create_location
from evohomeasync2.schemas import TCC_GET_LOC_STATUS
from evohomeasync2.schemas.config import factory_tcs, factory_time_zone
from evohomeasync2.schemas.const import (
    S2_GATEWAYS,
    S2_LOCATION_INFO,
    S2_SUPPORTS_DAYLIGHT_SAVING,
    S2_TEMPERATURE_CONTROL_SYSTEMS,
    S2_TIME_ZONE,
    S2_USE_DAYLIGHT_SAVE_SWITCHING,
)

from .conftest import ClientStub
from .const import TEST_DIR

if TYPE_CHECKING:
    import pytest

WORK_DIR = f"{TEST_DIR}/schemas_1"


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    def id_fnc(folder_path: Path) -> str:
        return folder_path.name

    folders = [
        p for p in Path(WORK_DIR).glob("*") if p.is_dir() and not p.name.startswith("_")
    ]
    metafunc.parametrize("folder", sorted(folders), ids=id_fnc)


# These schemas have camelCase keys, as per the vendor's schema
SCH_TCS_CONFIG: Final = factory_tcs()
SCH_TIME_ZONE: Final = factory_time_zone()


async def test_config_refresh(config: dict[str, Any], status: dict[str, Any]) -> None:
    """Test the loading a config, then an update_status() on top of that."""

    # hack because old JSON from HA's evohome integration didn't include this data
    if config[S2_LOCATION_INFO].get(S2_USE_DAYLIGHT_SAVE_SWITCHING) is None:
        config[S2_LOCATION_INFO][S2_USE_DAYLIGHT_SAVE_SWITCHING] = config[
            S2_LOCATION_INFO
        ][S2_TIME_ZONE][S2_SUPPORTS_DAYLIGHT_SAVING]

    config = convert_keys_to_snake_case(config)
    status = convert_keys_to_snake_case(status)

    # for this, we need snake_case keys
    loc = await create_location(ClientStub(), config)  # type: ignore[arg-type]
    loc._update_status(status)  # type: ignore[arg-type]


def test_config_schemas(config: dict[str, Any]) -> None:
    """Test the config schema for a location."""

    _ = SCH_TIME_ZONE(config[S2_LOCATION_INFO][S2_TIME_ZONE])

    for gwy_config in config[S2_GATEWAYS]:
        for tcs_config in gwy_config[S2_TEMPERATURE_CONTROL_SYSTEMS]:
            _ = SCH_TCS_CONFIG(tcs_config)


def test_status_schemas(status: dict[str, Any]) -> None:
    """Test the status schema for a location."""

    _ = TCC_GET_LOC_STATUS(status)



"""Tests for evohome-async."""

from __future__ import annotations

import json
import logging
import tempfile
from datetime import UTC, datetime as dt, timedelta as td
from pathlib import Path
from typing import TYPE_CHECKING, Any, Final, NotRequired, TypedDict

import aiofiles
import aiofiles.os

from evohomeasync.auth import (
    SZ_SESSION_ID,
    SZ_SESSION_ID_EXPIRES,
    AbstractSessionManager,
)
from evohomeasync2.auth import (
    SZ_ACCESS_TOKEN,
    SZ_ACCESS_TOKEN_EXPIRES,
    AbstractTokenManager,
)

if TYPE_CHECKING:
    from evohomeasync.auth import SessionIdEntryT
    from evohomeasync2.auth import AccessTokenEntryT


CACHE_FILE: Final = Path(tempfile.gettempdir()) / ".evo-cache.tmp"


class UserEntryT(TypedDict):
    access_token: NotRequired[AccessTokenEntryT]
    session_id: NotRequired[SessionIdEntryT]


CacheDataT = dict[str, UserEntryT]  # str is the client_id

"""
{
    "username@gmail.com": {
        "access_token": {
            "access_token": "iiuyz2-...",
            "access_token_expires": "2024-09-24T10:25:12+01:00",
            "refresh_token": "dfsadgf..."
            },
        "session_id": {
            "session_id": "94A76CB4-8BD4-4600-AAE4-...",
            "session_id_expires": "2024-09-24T10:25:12+01:00"
        }
    },
    "username@email.com": {}
  }
"""


class CredentialsManager(AbstractTokenManager, AbstractSessionManager):
    """A credentials manager that uses a file to cache the tokens."""

    def __init__(
        self, *args: Any, cache_file: Path | None = None, **kwargs: Any
    ) -> None:
        """Initialise the credentials manager (for access_token & session_id)."""

        # ensure default logger as we've merged the two ABCs
        kwargs["logger"] = kwargs.get("logger") or logging.getLogger(__name__)

        super().__init__(*args, **kwargs)

        self._cache_file: Final = cache_file

    @property
    def cache_file(self) -> str:
        """Return the token cache path."""
        return str(self._cache_file)

    @staticmethod
    def _clean_cache(old_cache: CacheDataT) -> CacheDataT:
        """Return a copy of a cache with any expired data removed."""

        new_cache: CacheDataT = {}

        dt_now = (dt.now(tz=UTC) + td(seconds=15)).isoformat()

        for user_id, entry in old_cache.items():
            user_data: UserEntryT = {}

            if (t := entry.get(SZ_ACCESS_TOKEN)) and t[
                SZ_ACCESS_TOKEN_EXPIRES
            ] > dt_now:
                user_data[SZ_ACCESS_TOKEN] = t

            # session_id is not used by evohomeasync2
            if (s := entry.get(SZ_SESSION_ID)) and s[SZ_SESSION_ID_EXPIRES] > dt_now:
                user_data[SZ_SESSION_ID] = s

            if user_data:
                new_cache[user_id] = user_data

        return new_cache  # could be Falsey

    async def _read_cache_from_file(self) -> CacheDataT:
        """Return a copy of the cache as read from file."""

        try:
            async with aiofiles.open(self.cache_file) as fp:
                content = await fp.read() or "{}"
        except FileNotFoundError:
            return {}

        cache: CacheDataT = json.loads(content)
        return cache

    async def _write_cache_to_file(self, cache: CacheDataT) -> None:
        """Write the supplied cache to file."""

        content = json.dumps(cache, indent=4)

        async with aiofiles.open(self.cache_file, "w") as fp:
            await fp.write(content)

    async def load_from_cache(self) -> None:
        """Load the user entry from the cache."""

        cache: CacheDataT = await self._read_cache_from_file()

        await self._load_access_token(cache=cache)
        await self._load_session_id(cache=cache)

    async def _load_access_token(self, cache: CacheDataT | None = None) -> None:
        """Load the (serialized) auth tokens from the cache."""

        cache = cache or await self._read_cache_from_file()

        entry: UserEntryT | None = cache.get(self.client_id)
        if not entry:
            return

        tokens: AccessTokenEntryT | None = entry.get(SZ_ACCESS_TOKEN)
        if not tokens:
            return

        # if not self._access_token:  # not needed as dt.min is sentinel for this
        if self._access_token_expires.isoformat() < tokens[SZ_ACCESS_TOKEN_EXPIRES]:
            self._import_access_token(tokens)

    async def _load_session_id(self, cache: CacheDataT | None = None) -> None:
        """Load the (serialized) session id from the cache."""

        cache = cache or await self._read_cache_from_file()

        entry: UserEntryT | None = cache.get(self.client_id)
        if not entry:
            return

        session: SessionIdEntryT | None = entry.get(SZ_SESSION_ID)
        if not session:
            return

        # if not self._session_id:  # not needed as dt.min is sentinel for this
        if self._session_id_expires.isoformat() < session[SZ_SESSION_ID_EXPIRES]:
            self._import_session_id(session)

    async def save_to_cache(self) -> None:
        """Save the user entry to the cache."""

        await self.save_access_token()
        await self.save_session_id()

    async def save_access_token(self) -> None:
        """Save the (serialized) access token to the cache.

        Includes the access token expiry datetime, and the refresh token.
        """

        cache: CacheDataT = await self._read_cache_from_file()

        if self.client_id not in cache:
            cache[self.client_id] = {}

        cache[self.client_id][SZ_ACCESS_TOKEN] = self._export_access_token()

        await self._write_cache_to_file(self._clean_cache(cache))

    async def save_session_id(self) -> None:
        """Save the (serialized) session id to the cache.

        Includes the session id expiry datetime.
        """

        cache: CacheDataT = await self._read_cache_from_file()

        if self.client_id not in cache:
            cache[self.client_id] = {}

        cache[self.client_id][SZ_SESSION_ID] = self._export_session_id()

        await self._write_cache_to_file(self._clean_cache(cache))



"""Validate the evohome-async v2 APIs (methods)."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

import evohomeasync2 as evo2
from _evohome.helpers import camel_to_snake
from evohomeasync2.schemas import SystemMode, factory_dhw_schedule, factory_zon_schedule
from evohomeasync2.schemas.const import S2_MODE
from tests.const import _DBG_USE_REAL_AIOHTTP

from . import faked_server as faked
from .common import skipif_auth_failed

if TYPE_CHECKING:
    from tests.conftest import EvohomeClientv2


def _get_dhw(evo: EvohomeClientv2) -> evo2.HotWater | None:
    """Return the DHW object of a TCS."""
    for loc in evo.locations:
        for gwy in loc.gateways:
            for tcs in gwy.systems:
                if tcs.hotwater:
                    return tcs.hotwater
    return None


def _get_zon(evo: EvohomeClientv2) -> evo2.Zone | None:
    """Return the Zone object of a TCS."""
    for loc in evo.locations:
        for gwy in loc.gateways:
            for tcs in gwy.systems:
                if tcs.zones:
                    return tcs.zones[0]
    return None


#######################################################################################


async def _test_usr_apis(evo: EvohomeClientv2) -> None:
    """Test User and Location methods.

    Includes: evo.user_account(), evo.installation() and loc.update() methods.
    """

    # STEP 1: retrieve config only: evo.user_account(), evo.installation()
    await evo.update(dont_update_status=True)

    assert evo2.main.SCH_USER_ACCOUNT(evo.user_account)
    assert evo2.main.SCH_USER_LOCATIONS(evo._user_locs)

    # STEP 2: GET /location/{loc.id}/status
    for loc in evo.locations:
        loc_status = await loc.update()
        assert evo2.Location.SCH_STATUS(loc_status)


async def _test_tcs_apis(evo: EvohomeClientv2) -> None:
    """Test ControlSystem methods.

    Includes tcs.update() and tcs.set_mode().
    Does not include tcs.get_schedules(), tcs.set_schedules().
    """

    # STEP 1: retrieve config only
    await evo.update(dont_update_status=False)

    # STEP 2: GET /temperatureControlSystem/{tcs.id}/status
    tcs = evo.locations[0].gateways[0].systems[0]

    # tcs_status = await tcs._update()
    # assert evo2.ControlSystem.SCH_STATUS(tcs_status)

    assert tcs.system_mode_status is not None
    mode = tcs.system_mode_status[S2_MODE]

    assert mode in SystemMode

    # STEP 3: PUT /temperatureControlSystem/{tcs.id}/mode
    await tcs.set_mode(SystemMode.AWAY)
    await evo.update()

    await tcs.set_mode(mode)


async def _test_dhw_apis(evo: EvohomeClientv2) -> None:
    """Test Hotwater methods.

    Includes dhw._update() and dhw.get_schedule().
    """

    # STEP 1: retrieve config only
    await evo.update(dont_update_status=True)

    if not (dhw := _get_dhw(evo)):
        pytest.skip("No DHW found in TCS")

    # STEP 2: GET /domesticHotWater/{dhw.id}/???
    dhw_status = await dhw._get_status()
    assert evo2.HotWater.SCH_STATUS(dhw_status)

    # STEP 2: GET /domesticHotWater/{dhw.id}/get_schedule
    schedule = await dhw.get_schedule()
    assert factory_dhw_schedule(camel_to_snake)({"daily_schedules": schedule})

    await dhw.set_schedule(schedule)


async def _test_zon_apis(evo: EvohomeClientv2) -> None:
    """Test Zone methods.

    Includes zon._update() and zon.get_schedule().
    """

    # STEP 1: retrieve config only
    await evo.update(dont_update_status=True)

    if not (zone := _get_zon(evo)):
        pytest.skip("No zones found in TCS")

    # STEP 2: GET /temperatureZone/{zon.id}/status
    zon_status = await zone._get_status()
    assert evo2.Zone.SCH_STATUS(zon_status)

    # STEP 2: GET /temperatureZone/{zon.id}/get_schedule
    if zone.id != faked.GHOST_ZONE_ID:
        schedule = await zone.get_schedule()
        assert factory_zon_schedule(camel_to_snake)({"daily_schedules": schedule})

        await zone.set_schedule(schedule)

    if zone := zone.tcs.zone_by_id.get(faked.GHOST_ZONE_ID):
        try:
            schedule = await zone.get_schedule()
        except evo2.InvalidScheduleError:
            pass
        else:
            pytest.fail("Did not raise expected exception")


#######################################################################################


@skipif_auth_failed
async def test_usr_apis(evohome_v2: EvohomeClientv2) -> None:
    """Test user_account() and installation()."""
    await _test_usr_apis(evohome_v2)


@skipif_auth_failed
async def test_tcs(evohome_v2: EvohomeClientv2) -> None:
    """Test set_mode() for TCS"""

    try:
        await _test_tcs_apis(evohome_v2)

    except NotImplementedError:  # TODO: implement
        if _DBG_USE_REAL_AIOHTTP:
            raise
        pytest.skip("Mocked server API not implemented")


@skipif_auth_failed
async def test_dhw_apis(evohome_v2: EvohomeClientv2) -> None:
    """Test get_schedule() and get_schedule()."""
    await _test_dhw_apis(evohome_v2)


@skipif_auth_failed
async def test_zon_apis(evohome_v2: EvohomeClientv2) -> None:
    """Test _update() for DHW/zone."""
    await _test_zon_apis(evohome_v2)



"""Provides handling of TCC gateways."""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Final, NoReturn

from _evohome.helpers import camel_to_snake

from .const import (
    SZ_GATEWAY_ID,
    SZ_GATEWAY_INFO,
    SZ_MAC,
    SZ_SYSTEM_ID,
    SZ_TEMPERATURE_CONTROL_SYSTEMS,
)
from .control_system import ControlSystem
from .schemas import factory_gwy_status
from .schemas.const import EntityType
from .zone import ActiveFaultsBase, EntityBase

if TYPE_CHECKING:
    import voluptuous as vol

    from . import Location
    from .schemas.typedefs import (
        EvoGwyConfigEntryT,
        EvoGwyConfigResponseT,
        EvoGwyStatusResponseT,
    )


class Gateway(ActiveFaultsBase, EntityBase):
    """Instance of a location's gateway."""

    SCH_STATUS: vol.Schema = factory_gwy_status(camel_to_snake)
    _TYPE = EntityType.GWY

    def __init__(self, location: Location, config: EvoGwyConfigResponseT) -> None:
        super().__init__(
            config[SZ_GATEWAY_INFO][SZ_GATEWAY_ID],
            location._auth,
            location._logger,
        )

        self.location = location  # parent
        #

        # children
        self.systems: list[ControlSystem] = []
        self.system_by_id: dict[str, ControlSystem] = {}  # tcs by id

        self._config: Final[EvoGwyConfigEntryT] = config[SZ_GATEWAY_INFO]  # type: ignore[misc]

        for tcs_entry in config[SZ_TEMPERATURE_CONTROL_SYSTEMS]:
            tcs = ControlSystem(self, tcs_entry)

            self.systems.append(tcs)
            self.system_by_id[tcs.id] = tcs

        self._status: EvoGwyStatusResponseT | None = None

    @property
    def config(self) -> EvoGwyConfigEntryT:
        """Return the latest config of the entity."""
        return self._config

    @property
    def status(self) -> EvoGwyStatusResponseT:
        """Return the latest status of the entity."""
        return super().status  # type: ignore[return-value]

    # Config attrs...

    @cached_property  # RENAMED val: was mac
    def mac_address(self) -> str:
        return self._config[SZ_MAC]

    # Status (state) attrs & methods...

    async def _get_status(self) -> NoReturn:
        """Get the latest state of the gateway and update its status attr.

        It is more efficient to call Location.update() as all descendants are updated
        with a single GET. Returns the raw JSON of the latest state.
        """

        raise NotImplementedError

    def _update_status(self, status: EvoGwyStatusResponseT) -> None:
        """Update the GWY's status and cascade to its descendants."""

        self._update_faults(status["active_faults"])

        # break the TypedDict into its parts (so, ignore[misc])...
        for tcs_status in status.pop(SZ_TEMPERATURE_CONTROL_SYSTEMS):  # type: ignore[misc]
            if tcs := self.system_by_id.get(tcs_status[SZ_SYSTEM_ID]):
                tcs._update_status(tcs_status)

            else:
                self._logger.warning(
                    f"{self}: system_id='{tcs_status[SZ_SYSTEM_ID]}' not known"
                    ", (has the gateway configuration been changed?)"
                )

        self._status = status



"""Tests for evohome-async - helper functions."""

from __future__ import annotations

import logging
from pathlib import Path
from typing import Final

from _evohome.const import HINT_BAD_CREDS, HINT_CHECK_NETWORK

TEST_DIR = Path(__file__).resolve().parent
FIXTURES_DIR = TEST_DIR / "fixtures"


MSG_INVALID_SESSION: Final = (
    "The session_id has been rejected (will re-authenticate): "
    "GET https://tccna.resideo.com/WebAPI/api/accountInfo: "
    '401 Unauthorized, response=[{"code": "Unauthorized", "message": "Unauthorized"}]'
)

MSG_INVALID_TOKEN: Final = (
    "The access_token has been rejected (will re-authenticate): "  # noqa: S105
    "GET https://tccna.resideo.com/WebAPI/emea/api/v1/userAccount: "
    '401 Unauthorized, response=[{"code": "Unauthorized", "message": "Unauthorized"}]'
)


LOG_01 = ("evohome_cli.auth", logging.DEBUG, "Fetching access_token...")
LOG_02 = ("evohome_cli.auth", logging.DEBUG, " - authenticating with the refresh_token")  # fmt: off
LOG_03 = ("evohome_cli.auth", logging.DEBUG, "Expired/invalid refresh_token")
LOG_04 = ("evohome_cli.auth", logging.DEBUG, " - authenticating with client_id/secret")  # fmt: off

LOG_21 = ("evohome_cli.auth", logging.DEBUG, "Fetching session_id...")
LOG_24 = ("evohome_cli.auth", logging.DEBUG, " - authenticating with client_id/secret")  # fmt: off

LOG_90 = ("evohome_cli.auth", logging.ERROR, HINT_BAD_CREDS)
LOG_99 = ("evohome_cli.auth", logging.ERROR, HINT_CHECK_NETWORK)

LOG_00 = ("evohomeasync", logging.WARNING, MSG_INVALID_SESSION)
LOG_09 = ("evohomeasync.auth", logging.ERROR, HINT_CHECK_NETWORK)

LOG_20 = ("evohomeasync2", logging.WARNING, MSG_INVALID_TOKEN)
LOG_29 = ("evohomeasync2.auth", logging.ERROR, HINT_CHECK_NETWORK)



"""Validate the handling of the vendor's v0 APIs (URLs) for Authentication.

This is used to:
a) document the RESTful API that is provided by the vendor
b) confirm the faked server (if any) is behaving as per a)

Testing is at HTTP request layer (e.g. GET).
Everything to/from the RESTful API is in camelCase (so those schemas are used).
"""

from __future__ import annotations

import random
import socket
import string
from http import HTTPStatus
from typing import TYPE_CHECKING
from urllib.parse import urlparse

import aiohttp
import pytest

from evohomeasync.auth import _APPLICATION_ID
from evohomeasync.schemas import (
    TCC_FAILURE_RESPONSE,
    TCC_GET_USR_INFO,
    TCC_POST_USR_SESSION,
)
from tests.const import (
    _DBG_TEST_CRED_URLS,
    _DBG_USE_REAL_AIOHTTP,
    HEADERS_BASE,
    HEADERS_CRED_V0,
    URL_BASE_V0,
    URL_CRED_V0,
)

if TYPE_CHECKING:
    from evohomeasync.schemas import (
        TccFailureResponseT,
        TccSessionResponseT,
        TccUserAccountInfoResponseT,
    )


async def _skip_if_too_many_requests(rsp: aiohttp.ClientResponse) -> None:
    if rsp.status != HTTPStatus.TOO_MANY_REQUESTS:  # 429
        return

    response: list[TccFailureResponseT] = await rsp.json()

    # the expected response for TOO_MANY_REQUESTS
    """
        [{
            "code": "TooManyRequests",
            "message": "Request count limitation exceeded, please try again later."
        }]
    """

    assert response[0]["code"] == "TooManyRequests"
    TCC_FAILURE_RESPONSE(response)

    pytest.skip("Too many requests")


@pytest.mark.skipif(not _DBG_USE_REAL_AIOHTTP, reason="requires vendor's webserver")
async def _test_bad0(  # invalid server certificate
    client_session: aiohttp.ClientSession,
) -> None:
    """Test authentication flow with an invalid server certificate."""

    #
    # TEST 0: invalid server certificate
    parsed_url = urlparse(URL_CRED_V0)
    ip_address = socket.gethostbyname(parsed_url.hostname)  # type: ignore[arg-type]
    url = f"{parsed_url.scheme}://{ip_address}{parsed_url.path}"  # path[:1] == "/"

    # aiohttp.client_exceptions.ClientConnectorCertificateError:
    # Cannot connect to host 135.224.164.118:443 ssl:True
    # [SSLCertVerificationError: ... certificate is not valid for ...]

    with pytest.raises(aiohttp.ClientConnectorCertificateError):
        await client_session.post(url, headers=HEADERS_CRED_V0, data={})


@pytest.mark.skipif(not _DBG_USE_REAL_AIOHTTP, reason="requires vendor's webserver")
async def test_bad1(  # bad credentials (client_id/secret)
    client_session: aiohttp.ClientSession,
) -> None:
    """Test authentication flow with bad credentials (client_id/secret)."""

    #
    # TEST 1: bad credentials (client_id/secret) -> HTTPStatus.UNAUTHORIZED
    data = {
        "applicationId": _APPLICATION_ID,
        "username": random.choice(string.ascii_letters),  # noqa: S311
        "password": "",
    }

    async with client_session.post(
        URL_CRED_V0, headers=HEADERS_CRED_V0, data=data
    ) as rsp:
        await _skip_if_too_many_requests(rsp)
        assert rsp.status == HTTPStatus.UNAUTHORIZED  # 401

        response: list[TccFailureResponseT] = await rsp.json()

        # the expected response for bad credentials
        """
            [{
                "code": "EmailOrPasswordIncorrect",
                "message": "The email or password provided is incorrect."
            }]
        """

    assert response[0]["code"] == "EmailOrPasswordIncorrect"
    TCC_FAILURE_RESPONSE(response)


@pytest.mark.skipif(not _DBG_USE_REAL_AIOHTTP, reason="requires vendor's webserver")
async def test_bad2(  # bad session id
    client_session: aiohttp.ClientSession,
) -> None:
    """Test authentication flow with an invalid/expired session id."""

    # pre-requisite data
    session_id = "bad session id " + random.choice(string.ascii_letters)  # noqa: S311

    #
    # TEST 9: bad session id -> HTTPStatus.UNAUTHORIZED
    url = f"{URL_BASE_V0}/accountInfo"
    headers = HEADERS_BASE | {"sessionId": session_id}

    async with client_session.get(url, headers=headers) as rsp:
        await _skip_if_too_many_requests(rsp)
        assert rsp.status == HTTPStatus.UNAUTHORIZED  # 401

        response: list[TccFailureResponseT] = await rsp.json()

        # the expected response for expired session id
        """
            [{
                "code": "Unauthorized",
                "message": "Unauthorized"
            }]
        """

    assert response[0]["code"] == "Unauthorized"
    TCC_FAILURE_RESPONSE(response)


@pytest.mark.skipif(not _DBG_USE_REAL_AIOHTTP, reason="requires vendor's webserver")
@pytest.mark.skipif(not _DBG_TEST_CRED_URLS, reason="may invalidate credentials cache")
async def test_good(  # good credentials
    client_session: aiohttp.ClientSession,
    credentials: tuple[str, str],
) -> None:
    """Test authentication flow (and authorization) with good credentials."""

    #
    # TEST 1: good credentials (client_id/secret) -> HTTPStatus.OK
    data = {
        "applicationId": _APPLICATION_ID,
        "username": credentials[0],
        "password": credentials[1],
    }

    async with client_session.post(
        URL_CRED_V0, headers=HEADERS_CRED_V0, data=data
    ) as rsp:
        await _skip_if_too_many_requests(rsp)
        assert rsp.status == HTTPStatus.OK  # 200

        user_auth: TccSessionResponseT = await rsp.json()

        # the expected response
        """
            {
                "sessionId": "A80FF794-C042-42BC-A63E-7A509C9AA6C9",
                "userInfo": {
                    "userID": 2263181,
                    "username": "username@email.com",
                    "firstname": "David",
                    "lastname": "Smith",
                    "streetAddress": "1 Main Street",
                    "city": "London",
                    "zipcode": "E1 1AA",
                    "country": "GB",
                    "telephone": "",
                    "userLanguage": "en-GB",
                    "isActivated": True,
                    "deviceCount": 0,
                    "tenantID": 5,
                    "securityQuestion1": "NotUsed",
                    "securityQuestion2": "NotUsed",
                    "securityQuestion3": "NotUsed",
                    "latestEulaAccepted": False
                }
            }
        """

    assert user_auth["userInfo"]["username"] == credentials[0]
    TCC_POST_USR_SESSION(user_auth)

    # #################################################################################

    # Check the session id by accessing a resource...
    session_id = user_auth["sessionId"]

    #
    # TEST 9: good session id -> HTTPStatus.OK
    url = f"{URL_BASE_V0}/accountInfo"
    headers = HEADERS_BASE | {"sessionId": session_id}

    async with client_session.get(url, headers=headers) as rsp:
        await _skip_if_too_many_requests(rsp)
        assert rsp.status == HTTPStatus.OK  # 200

        user_info: TccUserAccountInfoResponseT = await rsp.json()

        # the expected response
        """
            [
                {
                    "locationID": 1234567,
                    "name": "My Home",
                    "streetAddress": "1 Main Street",
                    "city": "London",
                    "state": "",
                    "country": "GB",
                    "zipcode": "E1 1AA",
                    "type": "Residential",
                    "hasStation": true,
                    "devices": [{}],  # list of ?DeviceResponse
                    "oneTouchButtons": [],
                    "weather": {
                        "condition": "Cloudy",
                        "temperature": 13.0,
                        "units": "Celsius",
                        "humidity": 96,
                        "phrase": "Cloudy"
                    },
                    "daylightSavingTimeEnabled": true,
                    "timeZone": {
                        "id": "GMT Standard Time",
                        "displayName": "(UTC+00:00) Dublin, Edinburgh, Lisbon, London",
                        "offsetMinutes": 0,
                        "currentOffsetMinutes": 0,
                        "usingDaylightSavingTime": true
                    },
                    "oneTouchActionsSuspended": false,
                    "isLocationOwner": true,
                    "locationOwnerID": 1234568,
                    "locationOwnerName": "David Smith",
                    "locationOwnerUserName": "username@email.com",
                    "canSearchForContractors": true,
                    "contractor": {
                        "info": {
                            "contractorID": 1839
                        },
                        "monitoring": {
                            "levelOfAccess": "Partial",
                            "contactPreferences": []
                        }
                    }
                }
            ]
        """

    assert user_auth["userInfo"]["username"] == credentials[0]
    TCC_GET_USR_INFO(user_info)



"""evohomeasync provides an async client for the v0 Resideo TCC API."""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from datetime import UTC, datetime as dt, timedelta as td
from typing import TYPE_CHECKING, Any, Final, TypedDict

import voluptuous as vol

from _evohome.auth import AbstractAuth
from _evohome.const import HEADERS_BASE, HEADERS_CRED, HINT_BAD_CREDS
from _evohome.credentials import CredentialsManagerBase
from _evohome.helpers import convert_keys_to_snake_case

from . import exceptions as exc
from .schemas import TCC_POST_USR_SESSION

if TYPE_CHECKING:
    import aiohttp
    from aiohttp.typedefs import StrOrURL

    from .schemas import EvoSessionDictT, EvoUserAccountDictT, TccSessionResponseT


# For API docs, enter this App ID on the following website under 'Session login':
#  - https://mytotalconnectcomfort.com/WebApi/Help/LogIn
_APPLICATION_ID: Final = "91db1612-73fd-4500-91b2-e63b069b185c"
#

SZ_SESSION_ID: Final = "session_id"
SZ_SESSION_ID_EXPIRES: Final = "session_id_expires"
SZ_USER_INFO: Final = "user_info"

# POST authentication url (i.e. /WebAPI/api/session)
URL_CRED: Final = "WebAPI/api/session"

# GET/PUT resource url (i.e. /WebAPI/api/...)
URL_BASE: Final = "WebAPI/api"


class SessionIdEntryT(TypedDict):
    session_id: str
    session_id_expires: str  # dt.isoformat()  # TZ-aware


class AbstractSessionManager(CredentialsManagerBase, ABC):
    """An ABC for managing the session id used for HTTP authentication."""

    _session_id: str
    _session_id_expires: dt
    _user_info: EvoUserAccountDictT | None

    def __init__(
        self,
        client_id: str,
        secret: str,
        websession: aiohttp.ClientSession,
        /,
        *,
        logger: logging.Logger | None = None,
        _hostname: str | None = None,
    ) -> None:
        """Initialise the session manager."""

        logger = logger or logging.getLogger(__name__)

        super().__init__(
            client_id, secret, websession, logger=logger, _hostname=_hostname
        )
        self._clear_session_id()  # initialise the attrs

    def _clear_session_id(self) -> None:
        """Clear the session id attrs (set to falsey state)."""

        self._session_id = ""
        self._session_id_expires = dt.min.replace(tzinfo=UTC)  # don't need local TZ

    @property
    def session_id(self) -> str:
        """Return the session id."""
        return self._session_id

    @property
    def session_id_expires(self) -> dt:
        """Return the expiration datetime of the session id."""
        return self._session_id_expires

    #
    #

    async def get_session_id(self) -> str:  # convenience wrapper
        """Return a valid session id.

        If required, fetch (and save) a new session id via the vendor's web API.
        """

        if not self.is_session_valid():  # although may be rejected for other reasons
            await self.fetch_session_id()
            await self.save_session_id()

        return self.session_id

    def is_session_valid(self) -> bool:
        """Return True if the session id is valid (the server may still reject it)."""
        return self._session_id_expires > dt.now(tz=UTC) + td(seconds=15)

    async def fetch_session_id(self) -> None:
        self._logger.debug("Fetching session_id...")

        self._logger.debug(" - authenticating with client_id/secret")

        credentials = {
            "applicationId": _APPLICATION_ID,
            "username": self._client_id,
            "password": self._secret,
        }

        try:  # check here if the user credentials are invalid...
            await self._fetch_session_id(credentials)

        except exc.AuthenticationFailedError as err:
            if "EmailOrPasswordIncorrect" not in err.message:
                raise
            self._logger.error(HINT_BAD_CREDS)  # noqa: TRY400
            raise exc.BadUserCredentialsError(f"{err}", status=err.status) from err

        self._was_authenticated = True

        # session_id is short-lived (but not safe to log self._user_info here)...
        self._logger.debug(f" - session_id = {self.session_id}")
        self._logger.debug(f" - session_id_expires = {self.session_id_expires}")

    async def _fetch_session_id(self, credentials: dict[str, str]) -> None:
        """Obtain an session id using the supplied credentials.

        The credentials are the user's client_id/secret.
        Will raise AuthenticationFailedError if unable to obtain a session id.
        """

        url = f"https://{self.hostname}/{URL_CRED}"

        response: TccSessionResponseT = await self._post_session_id_request(
            url,
            headers=HEADERS_CRED,
            data=credentials,  # NOTE: is camelCase
        )

        try:  # the dict _should_ be the expected schema...
            self._logger.debug(
                f"POST {url}: {TCC_POST_USR_SESSION(response)}"  # secrets are redacted via validator
            )

        except vol.Invalid as err:
            self._logger.warning(f"POST {url}: payload may be invalid: {err}")

        session: EvoSessionDictT = convert_keys_to_snake_case(response)  # type:ignore[assignment]

        try:
            self._session_id: str = session[SZ_SESSION_ID]
            self._session_id_expires = dt.now(tz=UTC) + td(minutes=15)
            self._user_info = session[SZ_USER_INFO]

        except (KeyError, TypeError) as err:
            raise exc.AuthenticationFailedError(
                f"Authenticator response is invalid: {response}: {err.__class__.__name__}: {err}"
            ) from err

        self._was_authenticated = True  # i.e. the credentials are valid

        if session.get("latest_eula_accepted"):
            self._logger.warning("The latest EULA has not been accepted by the user")

    async def _post_session_id_request(  # dev/test wrapper (also typing)
        self, url: StrOrURL, /, **kwargs: Any
    ) -> TccSessionResponseT:
        """Wrap the POST request to the vendor's TCC RESTful API."""
        return await self._post_request(url, **kwargs)  # type: ignore[return-value]

    @abstractmethod
    async def save_session_id(self) -> None:
        """Save the (serialized) session id (and expiry dtm).

        Should ideally confirm the session id is valid before saving.
        """

    def _import_session_id(self, session: SessionIdEntryT) -> None:
        """Extract the session id from a (serialized) dictionary."""

        self._session_id = session[SZ_SESSION_ID]
        self._session_id_expires = dt.fromisoformat(session[SZ_SESSION_ID_EXPIRES])

    def _export_session_id(self) -> SessionIdEntryT:
        """Convert the session id to a (serialized) dictionary."""

        return {
            SZ_SESSION_ID: self._session_id,
            SZ_SESSION_ID_EXPIRES: self._session_id_expires.isoformat(),
        }


class Auth(AbstractAuth):
    """A class to provide to access the v0 Resideo TCC API."""

    def __init__(
        self,
        session_manager: AbstractSessionManager,
        websession: aiohttp.ClientSession,
        /,
        *,
        logger: logging.Logger | None = None,
        _hostname: str | None = None,
    ) -> None:
        """A class for interacting with the v0 Resideo TCC API."""

        logger = logger or logging.getLogger(__name__)

        super().__init__(websession, logger=logger, _hostname=_hostname)

        self._session_id = session_manager.get_session_id
        self._url_base = f"https://{self.hostname}/{URL_BASE}"

    async def _headers(self, headers: dict[str, str] | None = None) -> dict[str, str]:
        """Ensure the authorization header has a valid session id."""

        headers = HEADERS_BASE | (headers or {})
        return headers | {
            "sessionId": await self._session_id(),
        }



"""Schema for vendor's TCC v2 API - for GET account of User, etc.

The convention for JSON keys is camelCase, but the API appears to be case-insensitive.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

import voluptuous as vol

from _evohome.helpers import noop, redact

from .const import (
    S2_CITY,
    S2_COUNTRY,
    S2_FIRSTNAME,
    S2_LANGUAGE,
    S2_LASTNAME,
    S2_POSTCODE,
    S2_STREET_ADDRESS,
    S2_USER_ID,
    S2_USERNAME,
)

if TYPE_CHECKING:
    from collections.abc import Callable


class TccOAuthTokenResponseT(TypedDict):
    """Typed dict for the OAuth authorization response schema.

    This schemas is snake_case, unlike the RESTful API which is camelCase.
    """

    access_token: str
    expires_in: int
    refresh_token: str
    token_type: str
    scope: str


def factory_post_oauth_token(fnc: Callable[[str], str] = noop) -> vol.Schema:
    """Factory for the OAuth authorization response schema."""

    # NOTE: These keys are always in snake_case

    return vol.Schema(
        {
            vol.Required("access_token"): vol.All(str, redact),
            vol.Required("expires_in"): vol.Range(min=1770, max=1800),  # usu. 179x
            vol.Required("refresh_token"): vol.All(str, redact),
            vol.Required("token_type"): str,
            vol.Optional("scope"): str,  # "EMEA-V1-Basic EMEA-V1-Anonymous"
        }
    )


class TccErrorResponseT(TypedDict):
    """Typed dict for error responses from the vendor servers."""

    error: str


def factory_error_response(fnc: Callable[[str], str] = noop) -> vol.Schema:
    """Factory for the error response schema."""

    return vol.Schema(
        {
            vol.Required("error"): str,
        },
        extra=vol.PREVENT_EXTRA,
    )


class TccUsrAccountResponseT(TypedDict):
    """Response to GET /userAccount"""

    userId: str  # '12345678' (i.e. str, not int)
    username: str  # user@mailbox.com
    firstname: str
    lastname: str
    streetAddress: str
    city: str
    postcode: str
    country: str  # UnitedKingdom
    language: str  # enGB


def factory_user_account(fnc: Callable[[str], str] = noop) -> vol.Schema:
    """Factory for the user account schema."""

    return vol.Schema(
        {
            vol.Required(fnc(S2_USER_ID)): str,
            vol.Required(fnc(S2_USERNAME)): vol.All(vol.Email(), redact),  # pyright: ignore[reportCallIssue]
            vol.Required(fnc(S2_FIRSTNAME)): str,
            vol.Required(fnc(S2_LASTNAME)): vol.All(str, redact),
            vol.Required(fnc(S2_STREET_ADDRESS)): vol.All(str, redact),
            vol.Required(fnc(S2_CITY)): vol.All(str, redact),
            vol.Required(fnc(S2_POSTCODE)): vol.All(str, redact),
            vol.Required(fnc(S2_COUNTRY)): str,
            vol.Required(fnc(S2_LANGUAGE)): str,
        },
        extra=vol.PREVENT_EXTRA,
    )


class TccFailureResponseT(TypedDict):
    """Typed dict for code/message responses from the vendor servers."""

    code: str
    message: str


def factory_status_response(fnc: Callable[[str], str] = noop) -> vol.Schema:
    """Factory for the error response schema."""

    entry_schema = vol.Schema(
        {
            vol.Required("code"): str,
            vol.Required("message"): str,
        },
        extra=vol.PREVENT_EXTRA,
    )

    return vol.Schema(vol.All([entry_schema], vol.Length(min=1)))


class TccTaskResponseT(TypedDict):
    """Typed dict for code/message responses from the vendor servers."""

    id: str  # {'id': '1668279943'}



"""Mocked vendor server for provision via a hacked aiohttp."""

from __future__ import annotations

import functools
import re
from http import HTTPMethod, HTTPStatus
from typing import TYPE_CHECKING, Any, TypeVar

import voluptuous as vol

from evohomeasync2.schemas import (
    TCC_GET_DHW_SCHEDULE,
    TCC_GET_ZON_SCHEDULE,
    const as sch,
)

from .const import (
    GHOST_ZONE_ID,
    MOCK_CRED_RESPONSE,
    MOCK_FULL_CONFIG,
    MOCK_LOCN_STATUS,
    MOCK_SCHEDULE_DHW,
    MOCK_SCHEDULE_ZONE,
    URL_BASE_V2 as URL_BASE,
    URL_CRED_V2 as URL_CRED,
    user_config_from_full_config as _user_config_from_full_config,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from .const import _bodyT, _methodT, _statusT, _urlT


def _dhw_id(url: _urlT) -> str:
    """Extract a DHW id from a URL."""
    return url.split(f"{sch.S2_DOMESTIC_HOT_WATER}/")[1].split("/")[0]


def _loc_id(url: _urlT) -> str:
    """Extract a Location id from a URL."""
    return url.split(f"{sch.S2_LOCATION}/")[1].split("/")[0]


def _tcs_id(url: _urlT) -> str:
    """Extract a TCS id from a URL."""
    return url.split(f"{sch.S2_TEMPERATURE_CONTROL_SYSTEM}/")[1].split("/")[0]


def _usr_id(url: _urlT) -> str:
    """Extract a User id from a URL."""
    return url.split("?userId=")[1].split("&")[0]


def _zon_id(url: _urlT) -> str:
    """Extract a Zone id from a URL."""
    return url.split(f"{sch.S2_TEMPERATURE_ZONE}/")[1].split("/")[0]


S = TypeVar("S", bound="FakedServerBase")


def validate_id_of_url(
    id_fnc: Callable[[_urlT], str],
) -> Callable[[Callable[[S], _bodyT | None]], Callable[[S], _bodyT]]:
    """Validate the id in the URL and set the status accordingly."""

    def decorator(
        fnc: Callable[[S], _bodyT | None],
    ) -> Callable[[S], _bodyT]:
        @functools.wraps(fnc)
        def wrapper(svr: S) -> _bodyT:
            if svr._method != HTTPMethod.GET:
                svr.status = HTTPStatus.METHOD_NOT_ALLOWED
                return {"message": "Method not allowed"}

            assert svr._url  # mypy
            try:
                id_: str = id_fnc(svr._url)
            except IndexError:
                svr.status = HTTPStatus.NOT_FOUND
                return {"message": "Not Found"}

            if not id_.isdigit():
                svr.status = HTTPStatus.BAD_REQUEST
                return [{"message": "Bad request"}]

            if result := fnc(svr):
                return result

            svr.status = HTTPStatus.UNAUTHORIZED
            return [{"message": "Unauthorized"}]

        return wrapper

    return decorator


class FakedServerBase:
    """Mocked vendor server for provision via a hacked aiohttp."""

    def __init__(
        self,
        full_config: dict,
        locn_status: dict,
        /,
        *,
        zone_schedule: dict | None = None,
        dhw_schedule: dict | None = None,
    ) -> None:
        self._full_config = full_config or MOCK_FULL_CONFIG
        self._locn_status = locn_status or MOCK_LOCN_STATUS
        self._zon_schedule = zone_schedule or MOCK_SCHEDULE_ZONE
        self._dhw_schedule = dhw_schedule or MOCK_SCHEDULE_DHW

        self._schedules: dict[str, dict] = {}
        self._user_config: dict[str, Any] = {}

        self.body: _bodyT | None = None
        self._method: _methodT | None = None
        self.status: _statusT | None = None
        self._url: _urlT | None = None

    def request(
        self, method: _methodT, url: _urlT, data: dict | str | None = None
    ) -> _bodyT:
        self._method = method
        self._url = url
        self._data = data

        self.body, self.status = None, None
        for pattern, fnc in REQUEST_MAP.items():
            if re.search(pattern, url):
                self.body = fnc(self)
                break
            # self.status = HTTPStatus.INTERNAL_SERVER_ERROR
        else:
            self.status = HTTPStatus.NOT_FOUND
            return """
                <html>
                    <head><title>Error</title></head>
                    <body><h1>Not Found</h1></body>
                </html>
            """

        if not self.status:
            self.status = HTTPStatus.OK if self.body else HTTPStatus.NOT_FOUND

        assert self.body is not None  # mypy
        return self.body


class FakedServerV0(FakedServerBase):
    """Mocked vendor server for provision of v0 URL responses."""

    def v0_session(self) -> _bodyT | None:
        raise NotImplementedError

    def v0_account_info(self) -> _bodyT | None:
        raise NotImplementedError

    def v0_locations(self) -> _bodyT | None:
        raise NotImplementedError

    def v0_evo_touch_systems(self) -> _bodyT | None:
        raise NotImplementedError

    def v0_heat_setpoint(self) -> _bodyT | None:
        raise NotImplementedError

    def v0_changeable_values(self) -> _bodyT | None:
        raise NotImplementedError


class FakedServerV2(FakedServerBase):
    """Mocked vendor server for provision of v2 URL responses."""

    def __init__(
        self,
        full_config: dict,
        locn_status: dict,
        /,
        *,
        zone_schedule: dict | None = None,
        dhw_schedule: dict | None = None,
    ) -> None:
        super().__init__(
            full_config,
            locn_status,
            zone_schedule=zone_schedule,
            dhw_schedule=dhw_schedule,
        )
        self._user_config = self._user_config_from_full_config(self._full_config)  # type: ignore[arg-type]

    def oauth_token(self) -> _bodyT | None:
        if self._method != HTTPMethod.POST:
            self.status = HTTPStatus.METHOD_NOT_ALLOWED
        elif self._url == URL_CRED:
            return MOCK_CRED_RESPONSE
        return None

    def usr_account(self) -> _bodyT:
        if self._method != HTTPMethod.GET:
            self.status = HTTPStatus.METHOD_NOT_ALLOWED
            return {"message": "Method not allowed"}

        if self._url == f"{URL_BASE}/userAccount":
            return self._user_config

        self.status = HTTPStatus.NOT_FOUND
        return {"message": "Not found"}

    @validate_id_of_url(_usr_id)
    def all_config(self) -> _bodyT | None:  # full_locn
        usr_id = _usr_id(self._url)  # type: ignore[arg-type]

        if self._user_config[sch.S2_USER_ID] == usr_id:
            return self._full_config
        return None

    def loc_config(self) -> _bodyT | None:
        raise NotImplementedError

    @validate_id_of_url(_loc_id)
    def loc_status(self) -> _bodyT | None:
        loc_id = _loc_id(self._url)  # type: ignore[arg-type]

        if self._locn_status[sch.S2_LOCATION_ID] == loc_id:
            return self._locn_status
        return None

    def tcs_mode(self) -> _bodyT | None:
        raise NotImplementedError

    @validate_id_of_url(_tcs_id)
    def tcs_status(self) -> _bodyT | None:
        tcs_id = _tcs_id(self._url)  # type: ignore[arg-type]

        for gwy in self._locn_status[sch.S2_GATEWAYS]:
            for tcs in gwy[sch.S2_TEMPERATURE_CONTROL_SYSTEMS]:  # type: ignore[index]
                if tcs[sch.S2_SYSTEM_ID] == tcs_id:
                    return tcs  # type: ignore[no-any-return]
        return None

    def zon_schedule(self) -> _bodyT | None:
        zon_id = _zon_id(self._url)  # type: ignore[arg-type]

        if self._method == HTTPMethod.GET:
            if zon_id == GHOST_ZONE_ID:
                self.status = HTTPStatus.BAD_REQUEST
                return [{"code": "ScheduleNotFound", "message": "Schedule not found."}]
            return self._schedules.get(zon_id, self._zon_schedule)

        if self._method != HTTPMethod.PUT:
            self.status = HTTPStatus.METHOD_NOT_ALLOWED
            return {"message": "Method not allowed"}

        if not isinstance(self._data, dict):
            self.status = HTTPStatus.BAD_REQUEST
            return [{"message": "Bad Request (invalid schedule: not a dict)"}]

        try:
            TCC_GET_ZON_SCHEDULE(self._data)
        except vol.Invalid:
            self.status = HTTPStatus.BAD_REQUEST
            return {"message": "Bad Request (invalid schedule: invalid schema)"}

        self._schedules[zon_id] = self._data
        return {"id": "1234567890"}

    def zon_mode(self) -> _bodyT | None:
        raise NotImplementedError

    @validate_id_of_url(_zon_id)
    def zon_status(self) -> _bodyT | None:
        zon_id = _zon_id(self._url)  # type: ignore[arg-type]

        for gwy in self._locn_status[sch.S2_GATEWAYS]:
            for tcs in gwy[sch.S2_TEMPERATURE_CONTROL_SYSTEMS]:  # type: ignore[index]
                for zone in tcs[sch.S2_ZONES]:
                    if zone[sch.S2_ZONE_ID] == zon_id:
                        return zone  # type: ignore[no-any-return]
        return None

    def dhw_schedule(self) -> _bodyT | None:
        dhw_id = _dhw_id(self._url)  # type: ignore[arg-type]

        if self._method == HTTPMethod.GET:
            return self._schedules.get(dhw_id, self._dhw_schedule)

        if self._method != HTTPMethod.PUT:
            self.status = HTTPStatus.METHOD_NOT_ALLOWED
            return {"message": "Method not allowed"}

        if not isinstance(self._data, dict):
            self.status = HTTPStatus.BAD_REQUEST
            return [{"message": "Bad Request (invalid schedule: not a dict)"}]

        try:
            TCC_GET_DHW_SCHEDULE(self._data)
        except vol.Invalid:
            self.status = HTTPStatus.BAD_REQUEST
            return {"message": "Bad Request (invalid schedule: invalid schema)"}

        self._schedules[dhw_id] = self._data
        return {"id": "1234567890"}

    @validate_id_of_url(_dhw_id)
    def dhw_status(self) -> _bodyT | None:
        dhw_id = _dhw_id(self._url)  # type: ignore[arg-type]

        for gwy in self._locn_status[sch.S2_GATEWAYS]:
            for tcs in gwy[sch.S2_TEMPERATURE_CONTROL_SYSTEMS]:  # type: ignore[index]
                if (dhw := tcs.get(sch.S2_DHW)) and dhw[sch.S2_DHW_ID] == dhw_id:
                    return dhw  # type: ignore[no-any-return]
        return None

    def dhw_mode(self) -> _bodyT | None:
        raise NotImplementedError

    @staticmethod
    def _user_config_from_full_config(full_config: list) -> dict:
        """Create a valid MOCK_USER_CONFIG from a MOCK_FULL_CONFIG."""
        return _user_config_from_full_config(full_config)


class FakedServer(FakedServerV2, FakedServerV0):
    """Mocked vendor server for provision via a hacked aiohttp."""


REQUEST_MAP_V0: dict[str, Callable] = {
    #
    r"/session": FakedServer.v0_session,  # authentication
    #
    r"/accountInfo$": FakedServer.v0_account_info,
    #
    r"/locations?userId=": FakedServer.v0_locations,
    #
    r"/evoTouchSystems?locationId=": FakedServer.v0_evo_touch_systems,
    #
    r"/devices/.*/thermostat/changeableValues/heatSetpoint": FakedServer.v0_heat_setpoint,
    r"/devices/.*/thermostat/changeableValues": FakedServer.v0_changeable_values,
}
REQUEST_MAP_V2: dict[str, Callable] = {
    #
    r"/Auth/OAuth/Token": FakedServer.oauth_token,  # authentication
    #
    r"/userAccount$": FakedServer.usr_account,
    #
    r"/location/installationInfo": FakedServer.all_config,
    r"/location/.*/installationInfo": FakedServer.loc_config,
    r"/location/.*/status": FakedServer.loc_status,
    #
    r"/temperatureControlSystem/.*/mode": FakedServer.tcs_mode,
    r"/temperatureControlSystem/.*/status": FakedServer.tcs_status,
    #
    r"/temperatureZone/.*/status": FakedServer.zon_status,
    r"/temperatureZone/.*/heatSetpoint": FakedServer.zon_mode,
    r"/temperatureZone/.*/schedule": FakedServer.zon_schedule,
    #
    r"/domesticHotWater/.*/status": FakedServer.dhw_status,
    r"/domesticHotWater/.*/state": FakedServer.dhw_mode,
    r"/domesticHotWater/.*/schedule": FakedServer.dhw_schedule,
}
REQUEST_MAP = REQUEST_MAP_V0 | REQUEST_MAP_V2



"""evohomeasync provides an async client for the v0 Resideo TCC API."""

from __future__ import annotations

import logging
from http import HTTPStatus
from typing import TYPE_CHECKING, Final

from _evohome.helpers import camel_to_snake

from . import exceptions as exc
from .auth import AbstractSessionManager, Auth
from .const import SZ_LOCATION_ID, SZ_USER_ID
from .entities import Location
from .schemas import factory_location_response_list, factory_user_account_info_response

if TYPE_CHECKING:
    import aiohttp

    from .schemas import EvoTcsInfoDictT, EvoUserAccountDictT

SCH_GET_ACCOUNT_INFO: Final = factory_user_account_info_response(camel_to_snake)
SCH_GET_ACCOUNT_LOCS: Final = factory_location_response_list(camel_to_snake)

_LOGGER = logging.getLogger(__name__.rpartition(".")[0])


class EvohomeClient:
    """Provide a client to access the Resideo TCC API."""

    _user_info: EvoUserAccountDictT | None = None
    _user_locs: list[EvoTcsInfoDictT] | None = None  # all locations of the user

    def __init__(
        self,
        session_manager: AbstractSessionManager,
        /,
        *,
        websession: aiohttp.ClientSession | None = None,
        debug: bool = False,
    ) -> None:
        """Construct the v0 EvohomeClient object."""

        self._logger = _LOGGER
        if debug:
            self._logger.setLevel(logging.DEBUG)
            self._logger.debug("Debug mode explicitly enabled via kwarg.")

        self._session_manager = session_manager
        self.auth = Auth(session_manager, websession or session_manager.websession)

        # self.devices: dict[_ZoneIdT, _DeviceDictT] = {}  # dhw or zone by id
        # self.named_devices: dict[_ZoneNameT, _DeviceDictT] = {}  # zone by name

        self._locations: list[Location] | None = None  # to preserve the order
        self._location_by_id: dict[str, Location] | None = None

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(auth='{self.auth}')"

    async def update(
        self,
        /,
        *,
        dont_update_status: bool = False,
        _reset_config: bool = False,  # for use by test suite
    ) -> list[EvoTcsInfoDictT]:
        """Retrieve the latest state of the user's locations.

        If required (or when `_reset_config` was true), first retrieves the user
        information & the configuration of all their locations.

        There is one API call for the user info, and a second for the config/status of
        all the user's locations.

        If `disable_status_update` is true, does not update the status of each location
        hierarchy (note: may have already retrieved the latest version of that data).
        """

        if _reset_config:
            self._user_info = None
            self._user_locs = None

            self._locations = None
            self._location_by_id = None

        if not dont_update_status:
            self._user_locs = None

        if self._user_locs is None:
            await self._get_config()

        assert self._user_locs is not None  # mypy

        if not dont_update_status:  # don't update status of location hierarchy
            assert self._location_by_id
            for loc_entry in self._user_locs:  # each entry is both config & status
                loc_id = str(loc_entry[SZ_LOCATION_ID])
                self._location_by_id[loc_id]._update_status(loc_entry)

        return self._user_locs

    async def _get_config(self) -> list[EvoTcsInfoDictT]:
        """Ensures the config of the user and their locations.

        If required, first retrieves the user information & installation configuration.
        """

        if self._user_info is None:  # will handle session_id rejection
            url = "accountInfo"
            try:
                self._user_info = await self.auth.get(url, schema=SCH_GET_ACCOUNT_INFO)  # type: ignore[assignment]

            except exc.ApiRequestFailedError as err:  # check if 401 - bad session_id
                if err.status != HTTPStatus.UNAUTHORIZED:  # 401
                    raise

                # as the accountInfo URL is open to all authenticated users, any 401 is
                # due the (albeit valid) session_id being rejected by the server

                self._logger.warning(
                    f"The session_id has been rejected (will re-authenticate): {err}"
                )

                self._session_manager._clear_session_id()
                self._user_info = await self.auth.get(url, schema=SCH_GET_ACCOUNT_INFO)  # type: ignore[assignment]

            assert self._user_info is not None  # mypy

        if self._user_locs is None:
            try:
                user_id = self._user_info[SZ_USER_ID]
            except (KeyError, TypeError) as err:
                raise exc.BadApiResponseError(
                    f"No user_id in user_info dict. Received: {self._user_info}"
                ) from err

            self._user_locs = await self.auth.get(
                f"locations?userId={user_id}&allData=True", schema=SCH_GET_ACCOUNT_LOCS
            )  # type: ignore[assignment]

            assert self._user_locs is not None  # mypy

        if self._locations is None:
            self._locations = []
            self._location_by_id = {}

            for loc_entry in self._user_locs:  # each entry is both config & status
                loc = Location(loc_entry[SZ_LOCATION_ID], loc_entry, self)
                self._locations.append(loc)
                self._location_by_id[loc.id] = loc

            #
            #

        return self._user_locs

    @property
    def user_account(self) -> EvoUserAccountDictT:
        """Return the information of the user account."""

        if self._user_info is None:
            raise exc.InvalidConfigError(
                "The account information is not (yet) available"
            )

        return self._user_info

    @property
    def locations(self) -> list[Location]:
        """Return the list of locations."""

        if self._user_locs is None:
            raise exc.InvalidConfigError(
                "The installation information is not (yet) available"
            )

        return self._locations  # type: ignore[return-value]

    @property
    def location_by_id(self) -> dict[str, Location]:
        """Return the list of locations."""

        if self._user_locs is None:
            raise exc.InvalidConfigError(
                "The installation information is not (yet) available"
            )

        return self._location_by_id  # type: ignore[return-value]

    # A significant majority of users will have exactly one TCS, thus for convenience...
    @property
    def tcs(self) -> Location:
        """If there is a single location/TCS, return it, or raise an exception.

        The majority of users will have only one location/TCS.
        """

        if not (locs := self.locations) or len(locs) != 1:
            raise exc.NoSingleTcsError(
                "There is not a single location (only) for this account"
            )

        return locs[0]



"""Tests for evohome-async - validate the schemas of vendor's RESTful JSON."""

from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import patch

from tests.conftest import EvohomeClientv2

from .conftest import FIXTURES_V2, auth_get

if TYPE_CHECKING:
    import pytest
    from freezegun.api import FrozenDateTimeFactory
    from syrupy import SnapshotAssertion

    from tests.conftest import CredentialsManager


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    folders = [
        p
        for p in Path(FIXTURES_V2).glob("*")
        if p.is_dir() and not p.name.startswith("_")
    ]
    metafunc.parametrize(
        "fixture_folder", sorted(folders), ids=(p.name for p in sorted(folders))
    )


async def test_system_warnings(
    credentials_manager: CredentialsManager,
    fixture_folder: Path,
    freezer: FrozenDateTimeFactory,
    caplog: pytest.LogCaptureFixture,
    snapshot: SnapshotAssertion,
) -> None:
    """Test the user account schema against the corresponding JSON."""

    freezer.move_to("2025-01-01T00:00:00+00:00")

    with patch("evohomeasync2.auth.Auth.get", auth_get(fixture_folder)):
        evo = EvohomeClientv2(credentials_manager)

        with caplog.at_level(logging.WARNING):
            await evo.update()

    assert caplog.record_tuples == snapshot



"""evohomeasync provides an async client for the v0 Resideo TCC API."""

from __future__ import annotations

import json
from functools import cached_property
from http import HTTPMethod
from typing import TYPE_CHECKING, Any, Final

import aiohttp

from . import exceptions as exc
from .auth import _payload
from .const import ERR_MSG_LOOKUP_BASE, HINT_CHECK_NETWORK, HOSTNAME

if TYPE_CHECKING:
    import logging

    from aiohttp.typedefs import StrOrURL


class CredentialsManagerBase:
    """A base class for managing the credentials used for HTTP authentication."""

    def __init__(
        self,
        client_id: str,
        secret: str,
        websession: aiohttp.ClientSession,
        /,
        *,
        logger: logging.Logger,
        _hostname: str | None = None,
    ) -> None:
        """Initialise the session manager."""

        self._client_id = client_id
        self._secret = secret
        self.websession: Final = websession

        self._logger = logger
        self._hostname: Final = _hostname or HOSTNAME

        self._was_authenticated = False  # True once credentials are proven valid

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return (
            f"{self.__class__.__name__}"
            f"(client_id='{self.client_id}, hostname='{self.hostname}')"
        )

    @cached_property
    def client_id(self) -> str:
        """Return the client id used for HTTP authentication."""
        return self._client_id

    @cached_property
    def hostname(self) -> str:
        """Return the hostname used for HTTP authentication."""
        return self._hostname

    async def _post_request(self, url: StrOrURL, /, **kwargs: Any) -> dict[str, Any]:
        """POST an authentication request and return the response (a dict).

        Will raise an exception if the authentication is not successful.
        """

        rsp: aiohttp.ClientResponse | None = None  # to prevent unbound error

        try:
            rsp = await self._request(HTTPMethod.POST, url, **kwargs)
            assert rsp is not None  # mypy hint

            await rsp.read()  # so we can use rsp.json()/rsp.text(), below
            rsp.raise_for_status()

            # can't assert content_length != 0 with aioresponses, so skip that check
            if rsp.content_type != "application/json":  # usu. "text/plain", "text/html"
                raise exc.AuthenticationFailedError(
                    f"Authenticator response is not JSON: {await _payload(rsp)}"
                )

            if (response := await rsp.json()) is None:  # an unanticipated edge-case
                raise exc.AuthenticationFailedError("Authenticator response is null")

        except (aiohttp.ContentTypeError, json.JSONDecodeError) as err:
            raise exc.AuthenticationFailedError(
                f"Authenticator response is not valid JSON: {await _payload(rsp)}"
            ) from err

        except aiohttp.ClientResponseError as err:
            # TODO: process payload and raise BadCredentialsError if code = EmailOrPasswordIncorrect
            if hint := ERR_MSG_LOOKUP_BASE.get(err.status):
                self._logger.error(hint)  # noqa: TRY400

            msg = f"{err.status} {err.message}, response={await _payload(rsp)}"

            raise exc.AuthenticationFailedError(
                f"Authenticator response is invalid: {msg}", status=err.status
            ) from err

        except aiohttp.ClientError as err:  # e.g. ClientConnectionError
            self._logger.error(HINT_CHECK_NETWORK)  # noqa: TRY400

            raise exc.AuthenticationFailedError(
                f"Authenticator response is invalid: {err}",
            ) from err

        else:
            return response  # type: ignore[no-any-return]

        finally:
            if rsp is not None:
                rsp.release()

    async def _request(  # dev/test wrapper
        self, method: HTTPMethod, url: StrOrURL, /, **kwargs: Any
    ) -> aiohttp.ClientResponse:
        """Wrap the request to the ClientSession (useful for dev/test)."""
        return await self.websession.request(method, url, **kwargs)



"""Mocked vendor RESTful API via a hacked aiohttp."""

from __future__ import annotations

from .aiohttp import ClientSession  # noqa: F401
from .const import GHOST_ZONE_ID  # noqa: F401
from .vendor import FakedServer  # noqa: F401



"""Tests for evohome-async - validate the schedule schemas."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from _evohome.helpers import convert_keys_to_snake_case
from evohomeasync2.schemas import TCC_GET_DHW_SCHEDULE, TCC_GET_ZON_SCHEDULE
from evohomeasync2.schemas.const import DayOfWeek
from evohomeasync2.zone import _find_switchpoints

from .conftest import JsonObjectType, load_fixture

if TYPE_CHECKING:
    from evohomeasync2.schemas.typedefs import DayOfWeekT

SCHEDULES_DIR = Path(__file__).parent / "schedules"


def schedule_file(filename: str) -> JsonObjectType:
    return load_fixture(SCHEDULES_DIR / filename)  # type: ignore[return-value]


SCHEDULE = convert_keys_to_snake_case(
    {
        "dailySchedules": [
            {
                "dayOfWeek": "Monday",
                "switchpoints": [
                    {"heatSetpoint": 23.2, "timeOfDay": "06:30:00"},
                    {"heatSetpoint": 18.0, "timeOfDay": "08:00:00"},
                    {"heatSetpoint": 19.1, "timeOfDay": "17:00:00"},
                    {"heatSetpoint": 14.9, "timeOfDay": "21:30:00"},
                ],
            },
            {
                "dayOfWeek": "Tuesday",
                "switchpoints": [
                    {"heatSetpoint": 19.2, "timeOfDay": "06:30:00"},
                    {"heatSetpoint": 18.2, "timeOfDay": "08:00:00"},
                    {"heatSetpoint": 19.3, "timeOfDay": "17:00:00"},
                    {"heatSetpoint": 14.9, "timeOfDay": "21:30:00"},
                ],
            },
            {
                "dayOfWeek": "Wednesday",
                "switchpoints": [
                    {"heatSetpoint": 19.1, "timeOfDay": "06:30:00"},
                    {"heatSetpoint": 18.0, "timeOfDay": "08:00:00"},
                    {"heatSetpoint": 19.1, "timeOfDay": "17:00:00"},
                    {"heatSetpoint": 14.9, "timeOfDay": "21:30:00"},
                ],
            },
            {
                "dayOfWeek": "Thursday",
                "switchpoints": [
                    {"heatSetpoint": 19.1, "timeOfDay": "06:30:00"},
                    {"heatSetpoint": 18.0, "timeOfDay": "08:00:00"},
                    {"heatSetpoint": 19.1, "timeOfDay": "17:00:00"},
                    {"heatSetpoint": 14.9, "timeOfDay": "21:30:00"},
                ],
            },
            {
                "dayOfWeek": "Friday",
                "switchpoints": [
                    {"heatSetpoint": 19.1, "timeOfDay": "06:30:00"},
                    {"heatSetpoint": 18.0, "timeOfDay": "08:00:00"},
                    {"heatSetpoint": 19.1, "timeOfDay": "17:00:00"},
                    {"heatSetpoint": 14.9, "timeOfDay": "21:30:00"},
                ],
            },
            {
                "dayOfWeek": "Saturday",
                "switchpoints": [
                    {"heatSetpoint": 19.1, "timeOfDay": "07:30:00"},
                    {"heatSetpoint": 18.5, "timeOfDay": "11:00:00"},
                    {"heatSetpoint": 19.1, "timeOfDay": "17:00:00"},
                    {"heatSetpoint": 14.9, "timeOfDay": "21:30:00"},
                ],
            },
            {
                "dayOfWeek": "Sunday",
                "switchpoints": [
                    {"heatSetpoint": 19.1, "timeOfDay": "07:30:00"},
                    {"heatSetpoint": 18.5, "timeOfDay": "11:00:00"},
                    {"heatSetpoint": 19.1, "timeOfDay": "17:00:00"},
                    {"heatSetpoint": 14.8, "timeOfDay": "21:30:00"},
                ],
            },
        ]
    }
)


def test_schema_schedule_dhw() -> None:
    """Test the schedule schema for dhw."""

    get_sched = schedule_file("schedule_dhw_get.json")
    _ = schedule_file("schedule_dhw_put.json")

    assert get_sched == TCC_GET_DHW_SCHEDULE(get_sched)
    # assert put_sched == TCC_PUT_DHW_SCHEDULE(put_sched)

    # assert put_sched == convert_to_put_schedule(get_sched)
    # assert get_sched == convert_to_get_schedule(put_sched)


def test_schema_schedule_zone() -> None:
    """Test the schedule schema for zones."""

    get_sched = schedule_file("schedule_zone_get.json")
    _ = schedule_file("schedule_zone_put.json")

    assert get_sched == TCC_GET_ZON_SCHEDULE(get_sched)
    # assert put_sched == TCC_PUT_ZON_SCHEDULE(put_sched)

    # assert put_sched == convert_to_put_schedule(get_sched)
    # assert get_sched == convert_to_get_schedule(put_sched)


def test_find_switchpoints() -> None:
    """Test the find_switchpoints method."""

    schedule: list[DayOfWeekT] = SCHEDULE["daily_schedules"]  # type: ignore[assignment]

    assert _find_switchpoints(schedule, DayOfWeek.MONDAY, "00:00:00") == (
        {"heat_setpoint": 14.8, "time_of_day": "21:30:00"},
        -1,
        {"heat_setpoint": 23.2, "time_of_day": "06:30:00"},
        0,
    )

    assert _find_switchpoints(schedule, DayOfWeek.TUESDAY, "07:59:59") == (
        {"heat_setpoint": 19.2, "time_of_day": "06:30:00"},
        0,
        {"heat_setpoint": 18.2, "time_of_day": "08:00:00"},
        0,
    )

    assert _find_switchpoints(schedule, DayOfWeek.TUESDAY, "08:00:00") == (
        {"heat_setpoint": 18.2, "time_of_day": "08:00:00"},
        0,
        {"heat_setpoint": 19.3, "time_of_day": "17:00:00"},
        0,
    )

    assert _find_switchpoints(schedule, DayOfWeek.TUESDAY, "08:00:01") == (
        {"heat_setpoint": 18.2, "time_of_day": "08:00:00"},
        0,
        {"heat_setpoint": 19.3, "time_of_day": "17:00:00"},
        0,
    )

    assert _find_switchpoints(schedule, DayOfWeek.SUNDAY, "23:59:59") == (
        {"heat_setpoint": 14.8, "time_of_day": "21:30:00"},
        0,
        {"heat_setpoint": 23.2, "time_of_day": "06:30:00"},
        1,
    )
